# -*- coding: utf-8 -*-
""" Utilitiy classes to define custom types.
"""

import six

from .._utils import cached_property, lazy
from ..exc import ScalarParsingError, ScalarSerializationError, UnknownEnumValue
from ..lang.parser import DIRECTIVE_LOCATIONS


def _evaluate_lazy_list(entries):
    _entries = lazy(entries)
    if not _entries:
        return []
    elif isinstance(_entries, (list, tuple)):
        return [lazy(entry) for entry in _entries]
    elif isinstance(_entries, dict):
        return [lazy(entry) for entry in _entries.values()]
    else:
        raise TypeError("Expected list or dict of items")


_UNDEF = object()


class Type(object):
    """ Base type class, all types used in a :class:`py_gql.schema.Schema`
    should be instances of this (or a subclass). """

    def __repr__(self):
        return str(self)

    def __eq__(self, lhs):
        return self is lhs or (
            isinstance(self, WrappingType)
            and self.__class__ == lhs.__class__
            # pylint: disable = no-member
            and self.type == lhs.type
        )

    def __hash__(self):
        return id(self)


class NamedType(Type):
    """ Named type. The name **must be unique** across a single
    :class:`py_gql.schema.Schema` instance.

    Attributes:
        name (str): Type name.
    """

    name = "ANONYMOUS"

    def __str__(self):
        return self.name


class WrappingType(Type):
    """ Types wrapping other types.

    Attributes:
        node (:class:`py_gql.lang.ast.Type`):
            Source node used when building type from the SDL

    Args:
        type_ (Lazy[py_gql.schema.Type]): Wrapped type.
        node (Optional[py_gql.lang.ast.Type]):
            Source node used when building type from the SDL
    """

    def __init__(self, type_, node=None):
        self._type = type_
        self.node = node

    @cached_property
    def type(self):
        """ :class:`py_gql.schema.Type`: Wrapped type """
        return lazy(self._type)


class NonNullType(WrappingType):
    """ Non nullable wrapping type.

    A non-null is a wrapping type which points to another type.
    Non-null types enforce that their values are never null and can ensure
    an error is raised if this ever occurs during a request. It is useful for
    fields which you can make a strong guarantee on non-nullability, for example
    usually the id field of a database row will never be null.

    Attributes:
        node (:class:`py_gql.lang.ast.NonNullType`):
            Source node used when building type from the SDL

    Args:
        type_ (Lazy[py_gql.schema.Type]): Wrapped type.
        node (Optional[py_gql.lang.ast.NonNullType]):
            Source node used when building type from the SDL
    """

    def __init__(self, type_, node=None):
        assert not isinstance(type_, NonNullType)
        super(NonNullType, self).__init__(type_, node=node)

    def __str__(self):
        return "%s!" % self.type


class ListType(WrappingType):
    """ List wrapping type.

    A list is a wrapping type which points to another type.
    Lists are often created within the context of defining the fields of
    an object type.

    Attributes:
        node (:class:`py_gql.lang.ast.ListType`):
            Source node used when building type from the SDL

    Args:
        type_ (Lazy[py_gql.schema.Type]):
            Type or callable returning the type (lazy / cyclic definitions).

        node (Optional[py_gql.lang.ast.ListType]):
            Source node used when building type from the SDL
    """

    def __str__(self):
        return "[%s]" % self.type


class InputField(object):
    """ Member of an :class:`py_gql.schema.InputObjectType`

    Args:
        name (str): Field name
        type_ (Lazy[py_gql.schema.Type]): Field type (must be input type)
        default_value (Optional[any]): Default value
        description (Optional[str]): Field description
        node (Optional[py_gql.lang.ast.InputValueDefinition]):
            Source node used when building type from the SDL

    Attributes:
        name (str): Field name
        description (Optional[str]): Field description
        has_default_value (bool): ``True`` if default value is set
        node (Optional[py_gql.lang.ast.InputValueDefinition]):
            Source node used when building type from the SDL
    """

    # Yikes! Didn't find a better way to differentiate None as value and no
    # value in arguments... at least it's not exposed to callers.
    # Maybe we could wrap default value in a singleton type ?
    def __init__(
        self, name, type_, default_value=_UNDEF, description=None, node=None
    ):
        self.name = name
        self._type = type_
        self._default_value = default_value
        self.description = description
        self.has_default_value = self._default_value is not _UNDEF
        self.node = node

    @property
    def default_value(self):
        """ Default value if it was set """
        if self._default_value is _UNDEF:
            raise AttributeError("No default value")
        return self._default_value

    @cached_property
    def type(self):
        """ py_gql.schema.Type: Field type """
        return lazy(self._type)

    @cached_property
    def required(self):
        """ bool: Whether this field is required (non nullable and does
        not have a default value) """
        return (
            isinstance(self.type, NonNullType) and self._default_value is _UNDEF
        )

    def __str__(self):
        return "InputField(%s: %s)" % (self.name, self.type)


class InputObjectType(NamedType):
    """ Input Object Type Definition

    An input object defines a structured collection of fields which may be
    supplied to a field argument or a directive.

    Args:
        name (str): Type name
        fields (Lazy[List[py_gql.schema.InputField]]): Fields
        description (Optional[str]): Type description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    Attributes:
        name (str): Type name
        description (Optional[str]): Directive description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    """

    def __init__(self, name, fields, description=None, nodes=None):
        self.name = name
        self.description = description
        self._fields = fields
        self.nodes = [] if nodes is None else nodes

    @cached_property
    def fields(self):
        """ List[py_gql.schema.InputField]: Object fields """
        return _evaluate_lazy_list(self._fields)

    @cached_property
    def field_map(self):
        """ Dict[str, py_gql.schema.InputField]: Object fields map """
        return {f.name: f for f in self.fields}


class EnumValue(object):
    """ Enum value definition.

    Args:
        name (str): Name of the value
        value (Optional[any]): Python value.
            Defaults to ``name`` if unset, must be hashable to support reverse
            lookups
        deprecation_reason (Optional[str]):
            If set, the field will be marked as deprecated and include this as
            the reason
        description (Optional[str]): Enum value description
        node (Optional[py_gql.lang.ast.EnumValueDefinition]):
            Source node used when building type from the SDL

    Attributes:
        name (str): Enum value name
        value: Enum value value
        description (Optional[str]): Enum value description
        deprecation_reason (Optional[str]):
        deprecated (bool):
        node (Optional[py_gql.lang.ast.FieldDefinition]):
            Source node used when building type from the SDL
    """

    @classmethod
    def from_def(cls, definition):
        """ Create an enum value from various source objects.

        This supports existing `EnumValue` instances, strings,
        (name, value) tuples and dictionnaries matching the signature of
        `EnumValue.__init__`.
        """
        if isinstance(definition, cls):
            return definition
        elif isinstance(definition, str):
            return cls(definition, definition)
        elif isinstance(definition, tuple):
            name, value = definition
            return cls(name, value)
        elif isinstance(definition, dict):
            return cls(**definition)
        else:
            raise TypeError("Invalid enum value definition %r" % definition)

    def __init__(
        self,
        name,
        value=_UNDEF,
        deprecation_reason=None,
        description=None,
        node=None,
    ):
        assert name not in ("true", "false", "null")
        self.name = name
        self.value = value if value is not _UNDEF else name
        self.description = description
        self.deprecated = bool(deprecation_reason)
        self.deprecation_reason = deprecation_reason
        self.node = node

    def __str__(self):
        return self.name


class EnumType(NamedType):
    """ Enum Type Definition

    Some leaf values of requests and input values are Enums. GraphQL serializes
    Enum values as strings, however internally Enums can be represented by any
    kind of type, often integers.

    Warning:
        Enum values must be hashable to provide reverse lookup
        capabilities when coercing python values into enum values.

    Args:
        name (str): Enum name
        values (list): List of enum value definition support by
            `EnumValue.from_def`
        description (Optional[str]): Enum description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    Attributes:
        name (str): Enum name
        values (Dict[str, py_gql.schema.EnumValue]):
            Values by name
        reverse_values (Dict[any, py_gql.schema.EnumValue]):
            Values by value
        description (Optional[str]): Enum description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    """

    def __init__(self, name, values, description=None, nodes=None):
        self.name = name
        self.description = description
        self.nodes = [] if nodes is None else nodes
        self.values = []
        self._values = {}
        self._reverse_values = {}
        for v in values:
            ev = EnumValue.from_def(v)
            assert ev.name not in self._values, (
                "Duplicate enum value %s" % ev.name
            )
            self.values.append(ev)
            self._reverse_values[ev.value] = self._values[ev.name] = ev

    def get_value(self, name):
        """ Extract the value for a given name.

        Args:
            name (str): Name of the value to extract

        Returns:
            :class:`py_gql.schema.EnumValue`: The corresponding EnumValue

        Raises:
            :class:`~py_gql.exc.UnknownEnumValue` when the name is unknown
        """
        try:
            return self._values[name].value
        except KeyError:
            raise UnknownEnumValue(
                "Invalid name %s for enum %s" % (name, self.name)
            )

    def get_name(self, value):
        """ Extract the name for a given value.

        Args:
            value (any): Value of the value to extract, must be hashable

        Returns:
            :class:`py_gql.schema.EnumValue`: The corresponding EnumValue

        Raises:
            :class:`~py_gql.exc.UnknownEnumValue` when the value is unknown
        """
        try:
            return self._reverse_values[value].name
        except KeyError:
            raise UnknownEnumValue(
                "Invalid value %r for enum %s" % (value, self.name)
            )


class ScalarType(NamedType):
    """ Scalar Type Definition

    The leaf values of any request and input values to arguments are
    Scalars (or Enums) and are defined with a name and a series of functions
    used to parse input from ast or variables and to ensure validity.

    Args:
        name (str): Type name

        serialize (callable): Type serializer
            This function receives Python value and must output JSON scalars
            (that is string, number, boolean and null).
            Raise :class:`~py_gql.exc.ScalarSerializationError`,
            :py:class:`ValueError` or :py:class:`TypeError` to signify that the
            value cannot be serialized.

        parse (callable): Type de-serializer
            This function receives JSON scalars and outputs Python values.
            Raise :class:`~py_gql.exc.ScalarParsingError`, :py:class:`ValueError`
            or :py:class:`TypeError` to signify that the value cannot be parsed.

        parse_literal (Optional[callable]): Type de-serializer for value nodes
            This function receives a :class:`py_gql.lang.ast.Value`
            and outputs Python values.
            Raise :class:`~py_gql.exc.ScalarParsingError`, :py:class:`ValueError`
            or :py:class:`TypeError` to signify that the value cannot be parsed.
            If unset, the ``parse`` argument is used and gets passed the
            ``value`` attribute of the node.

        description (Optional[str]): Type description

        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    Attributes:
        name (str): Type name
        description (Optional[str]): Type description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    """

    def __init__(
        self,
        name,
        serialize,
        parse,
        parse_literal=None,
        description=None,
        nodes=None,
    ):
        self.name = name
        self.description = description
        assert callable(serialize)
        self._serialize = serialize
        assert callable(parse)
        self._parse = parse
        assert parse_literal is None or callable(parse_literal)
        self._parse_literal = parse_literal
        self.nodes = [] if nodes is None else nodes

    def serialize(self, value):
        """ Transform a Python value in a JSON serializable one.

        Args:
            value: Python level value

        Returns:
            JSON scalar
        """
        try:
            return self._serialize(value)
        except (ValueError, TypeError) as err:
            six.raise_from(ScalarSerializationError(str(err)), err)

    def parse(self, value):
        """ Transform a GraphQL value in a valid Python value

        Args:
            value: JSON scalar

        Returns:
            Python level value
        """
        try:
            return self._parse(value)
        except (ValueError, TypeError) as err:
            six.raise_from(ScalarParsingError(str(err)), err)

    def parse_literal(self, node, variables=None):
        """ Transform an AST node in a valid Python value

        Args:
            value (py_gql.lang.ast.Value): Parse node

        Returns:
            Python level value
        """
        try:
            if (
                hasattr(self, "_parse_literal")
                and self._parse_literal is not None
            ):
                return self._parse_literal(node, variables or {})
            else:
                return self.parse(node.value)
        except (ValueError, TypeError) as err:
            six.raise_from(ScalarParsingError(str(err), [node]), err)


class Argument(object):
    """ Field or Directive argument definition.

    Warning:
        As ``None`` is a valid default value, in order to define an
        argument wihtout any default value, the ``default_value`` argument
        must be omitted.

    Args:
        name (str): Argument name
        type_ (Lazy[py_gql.schema.Type]): Argument type (must be input type)
        default_value (Optional[any]): Default value
        description (Optional[str]): Argument description
        node (Optional[py_gql.lang.ast.InputValueDefinition]):
            Source node used when building type from the SDL

    Attributes:
        name (str): Argument name
        description (Optional[str]): Argument description
        has_default_value (bool): ``True`` if default value is set
        node (Optional[py_gql.lang.ast.InputValueDefinition]):
            Source node used when building type from the SDL
    """

    # Yikes! Didn't find a better way to differentiate None as value and no
    # value in arguments... at least it's not exposed to callers.
    # Maybe we could wrap default value in a singleton type ?
    def __init__(
        self, name, type_, default_value=_UNDEF, description=None, node=None
    ):
        self.name = name
        self._type = type_
        self._default_value = default_value
        self.description = description
        self.has_default_value = self._default_value is not _UNDEF
        self.node = node

    @property
    def default_value(self):
        """ Default value if it was set """
        if self._default_value is _UNDEF:
            raise AttributeError("No default value")
        return self._default_value

    @cached_property
    def type(self):
        """ py_gql.schema.Type: Argument type """
        return lazy(self._type)

    @cached_property
    def required(self):
        """ bool: Whether this argument is required (non nullable and does
        not have a default value) """
        return (
            isinstance(self.type, NonNullType) and self._default_value is _UNDEF
        )

    def __str__(self):
        return "Argument(%s: %s)" % (self.name, self.type)


class Field(object):
    """ Member of an :class:`py_gql.schema.ObjectType`.

    Args:
        name (str): Field name
        type_ (Lazy[py_gql.schema.Type]): Field type (must be output type)
        args (Lazy[List[py_gql.schema.Argument]]): Field arguments
        description (Optional[str]): Field description
        deprecation_reason (Optional[str]):
            If set, the field will be marked as deprecated and include this as
            the reason
        resolve (callable):
            Resolver function. If not set, :func:`py_gql.utilities.default_resolver`
            will be used during execution
        node (Optional[py_gql.lang.ast.FieldDefinition]):
            Source node used when building type from the SDL

    Attributes:
        name (str): Field name
        description (Optional[str]): Field description
        deprecation_reason (Optional[str]):
        deprecated (bool):
        resolve (Optional[callable]): Resolver function
        node (Optional[py_gql.lang.ast.FieldDefinition]):
            Source node used when building type from the SDL
    """

    def __init__(
        self,
        name,
        type_,
        args=None,
        description=None,
        deprecation_reason=None,
        resolve=None,
        node=None,
    ):
        assert resolve is None or callable(resolve)

        self.name = name
        self._type = type_
        self.description = description
        self.deprecated = bool(deprecation_reason)
        self.deprecation_reason = deprecation_reason
        self.resolve = resolve
        self._args = args
        self.node = node

    @cached_property
    def type(self):
        """ py_gql.schema.Type: Field type """
        return lazy(self._type)

    @cached_property
    def args(self):
        """ List[py_gql.schema.Argument]: Field arguments """
        return _evaluate_lazy_list(self._args)

    arguments = args

    @cached_property
    def arg_map(self):
        """ Dict[str, py_gql.schema.Argument]: Field arguments map """
        return {arg.name: arg for arg in self.args}

    def __str__(self):
        return "Field(%s: %s)" % (self.name, self.type)


class ObjectType(NamedType):
    """ Object Type Definition

    Almost all of the GraphQL types you define will be object types. Object
    types have a name, but most importantly describe their fields.

    Args:
        name (str): Type name
        fields (Lazy[List[py_gql.schema.Field]]): Fields
        interfaces (Lazy[List[py_gql.schema.InterfaceType]]):
            Implemented interfaces
        is_type_of:
            Either ``None``, a callable or a class used to identify an
            object's interface.
        description (Optional[str]): Type description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    Attributes:
        name (str): Type name
        is_type_of (Optional[callable]): Type resolver
        description (Optional[str]): Directive description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL
    """

    def __init__(
        self,
        name,
        fields,
        interfaces=None,
        is_type_of=None,
        description=None,
        nodes=None,
    ):
        self.name = name
        self.description = description
        self._fields = fields
        self._interfaces = interfaces
        self.nodes = [] if nodes is None else nodes

        assert is_type_of is None or callable(is_type_of)
        if isinstance(is_type_of, type):
            self.is_type_of = lambda v, **_: isinstance(v, is_type_of)
        else:
            self.is_type_of = is_type_of

    @cached_property
    def interfaces(self):
        """ List[py_gql.schema.InterfaceType]: Implemented interfaces """
        return _evaluate_lazy_list(self._interfaces)

    @cached_property
    def fields(self):
        """ List[py_gql.schema.Field]: Object fields """
        return _evaluate_lazy_list(self._fields)

    @cached_property
    def field_map(self):
        """ Dict[str, py_gql.schema.Field]: Object fields map """
        return {f.name: f for f in self.fields}


class InterfaceType(NamedType):
    """ Interface Type Definition

    When a field can return one of a heterogeneous set of types, a Interface
    type is used to describe what types are possible, what fields are in
    common across all types, as well as a function to determine which type
    is actually used when the field is resolved.

    Args:
        name (str): Type name
        fields (Lazy[List[py_gql.schema.Field]]): Fields
        resolve_type (Optional[callable]): Type resolver
        description (Optional[str]): Type description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    Attributes:
        name (str): Type name
        resolve_type (Optional[callable]): Type resolver
        description (Optional[str]): Directive description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL
    """

    def __init__(
        self, name, fields, resolve_type=None, description=None, nodes=None
    ):
        self.name = name
        self.description = description
        self._fields = fields
        self.nodes = [] if nodes is None else nodes

        assert resolve_type is None or callable(resolve_type)
        self.resolve_type = resolve_type

    @cached_property
    def fields(self):
        """ List[py_gql.schema.Field]: Interface fields """
        return _evaluate_lazy_list(self._fields)

    @cached_property
    def field_map(self):
        """ Dict[str, py_gql.schema.Field]: Interface fields map """
        return {f.name: f for f in self.fields}


class UnionType(NamedType):
    """ Union Type Definition

    When a field can return one of a heterogeneous set of types, a Union type
    is used to describe what types are possible as well as providing a function
    to determine which type is actually used when the field is resolved.

    Args:
        name (str): Type name
        types (Lazy[List[py_gql.schema.ObjectType]]): Member types
        resolve_type (Optional[callable]): Type resolver
        description (Optional[str]): Type description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    Attributes:
        name (str): Type name
        resolve_type (Optional[callable]): Type resolver
        description (Optional[str]): Directive description
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Source nodes used when building type from the SDL

    """

    def __init__(
        self, name, types, resolve_type=None, description=None, nodes=None
    ):
        """
        :type name: str
        :type types: callable|[Type]|dict(str, Type)
        :type description: str
        """
        self.name = name
        self.description = description
        self._types = types
        self.nodes = [] if nodes is None else nodes

        assert resolve_type is None or callable(resolve_type)
        self.resolve_type = resolve_type

    @cached_property
    def types(self):
        """ List[py_gql.schema.ObjectType]: Member types """
        return _evaluate_lazy_list(self._types)


class Directive(Type):
    """ Directive definition

    Directives are used by the GraphQL runtime as a way of modifying
    execution behavior. Type system creators will usually not create
    these directly.

    Args:
        name (str): Directive name
        locations (List[str]): Possible locations for that directive
        args (List[py_gql.schena.Argument]): Argument definitions
        description (Optional[str]): Directive description
        node (Optional[py_gql.lang.ast.DirectiveDefinition]):
            Source node used when building type from the SDL.

    Attributes:
        name (str): Directive name
        locations (List[str]): Possible locations for that directive
        args (List[py_gql.schena.Argument]): Argument definitions
        arguments (List[py_gql.schena.Argument]): Argument definitions
        arg_map: (Dict[str, py_gql.schena.Argument]): ``arg name -> arg definition``
        description (Optional[str]): Directive description
        node (Optional[py_gql.lang.ast.DirectiveDefinition]):
            Source node used when building type from the SDL.
    """

    def __init__(self, name, locations, args=None, description=None, node=None):
        assert locations and all(
            [loc in DIRECTIVE_LOCATIONS for loc in locations]
        )
        self.name = name
        self.description = description
        self.locations = locations
        self.arguments = self.args = args or []
        self.arg_map = {arg.name: arg for arg in self.args}
        self.node = node

    def __str__(self):
        return "@%s" % self.name


def is_input_type(type_):
    """ These types may be used as input types for arguments and directives.

    Args:
        type_ (py_gql.schema.Type): Type under test

    Returns:
        bool:
    """
    return isinstance(
        unwrap_type(type_), (ScalarType, EnumType, InputObjectType)
    )


def is_output_type(type_):
    """ These types may be used as output types as the result of fields.

    Args:
        type_ (py_gql.schema.Type): Type under test

    Returns:
        bool:
    """
    return isinstance(
        unwrap_type(type_),
        (ScalarType, EnumType, ObjectType, InterfaceType, UnionType),
    )


def is_leaf_type(type_):
    """  These types may describe types which may be leaf values.

    Args:
        type_ (py_gql.schema.Type): Type under test

    Returns:
        bool:
    """
    return isinstance(type_, (ScalarType, EnumType))


def is_composite_type(type_):
    """ These types may describe the parent context of a selection set.

    Args:
        type_ (py_gql.schema.Type): Type under test

    Returns:
        bool:
    """
    return isinstance(type_, (ObjectType, InterfaceType, UnionType))


def is_abstract_type(type_):
    """ These types may describe the parent context of a selection set.

    Args:
        type_ (py_gql.schema.Type): Type under test

    Returns:
        bool:
    """
    return isinstance(type_, (InterfaceType, UnionType))


def unwrap_type(type_):
    """ Recursively extract type for a potentially wrapping type like
    :class:`ListType` or :class:`NonNullType`.

    Args:
        type_ (py_gql.schema.Type): Potentially wrapped type

    Returns:
        Unwrapped type

    >>> from py_gql.schema import Int, NonNullType, ListType
    >>> unwrap_type(NonNullType(ListType(NonNullType(Int)))) is Int
    True
    """
    if isinstance(type_, WrappingType):
        return unwrap_type(type_.type)
    return type_


def nullable_type(type_):
    """ Extract nullable type from a potentially non nulllable one.

    Args:
        type_ (py_gql.schema.Type): Potentially non-nullable type

    Returns:
        py_gql.schema.Type:  Nullable type

    >>> from py_gql.schema import Int, NonNullType
    >>> unwrap_type(NonNullType(Int)) is Int
    True

    >>> unwrap_type(Int) is Int
    True
    """
    if isinstance(type_, NonNullType):
        return type_.type
    return type_



# -*- coding: utf-8 -*-
""" Parser tests.
Ported from: https://github.com/graphql/graphql-js/blob/master/src/language/__tests__/schema-parser-test.js  # noqa: E501
on revision 8d1ae25de444a9b544a1fdc98e138ae91b77057c.
"""
# Test naming is not very 'pythonic' as I tried to keep a close match with the
# original ones for easy reference. Others were kept while they were specific
# to the implementation and are kept as skipped.


import pytest

from py_gql.exc import UnexpectedEOF, UnexpectedToken
from py_gql.lang import ast as _ast
from py_gql.lang.parser import parse


# Comparing dicts will result in better assertion diffs from pytest.
def assert_node_equal(ref, expected):
    # import json
    # print(json.dumps(_ast.node_to_dict(ref), sort_keys=True, indent=4))
    # print(json.dumps(_ast.node_to_dict(expected), sort_keys=True, indent=4))
    assert _ast.node_to_dict(ref) == _ast.node_to_dict(expected)


# A few syntactic sugar helpers
def _name(loc, value):
    return _ast.Name(loc=loc, value=value)


def _type(loc, value):
    return _ast.NamedType(loc=loc, name=_ast.Name(loc=loc, value=value))


def _field(loc, name, type_, args=[]):  # noqa : B006
    return _ast.FieldDefinition(
        loc=loc, name=name, type=type_, arguments=args, directives=[]
    )


def _input(loc, name, type_, default_value=None):
    return _ast.InputValueDefinition(
        loc=loc,
        name=name,
        type=type_,
        default_value=default_value,
        directives=[],
    )


def _doc(loc, defs):
    return _ast.Document(loc=loc, definitions=defs)


def test_it_parses_simple_type():
    body = u"""
type Hello {
  world: String
}"""
    assert_node_equal(
        parse(body),
        _ast.Document(
            loc=(0, 31),
            definitions=[
                _ast.ObjectTypeDefinition(
                    loc=(1, 31),
                    name=_name((6, 11), "Hello"),
                    interfaces=[],
                    directives=[],
                    fields=[
                        _field(
                            (16, 29),
                            _name((16, 21), "world"),
                            _type((23, 29), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_type_with_description_string():
    assert_node_equal(
        parse(
            u"""
"Description"
type Hello {
  world: String
}"""
        ),
        _ast.Document(
            loc=(0, 45),
            definitions=[
                _ast.ObjectTypeDefinition(
                    loc=(1, 45),
                    name=_name((20, 25), "Hello"),
                    interfaces=[],
                    directives=[],
                    description=_ast.StringValue(
                        loc=(1, 14), value="Description"
                    ),
                    fields=[
                        _field(
                            (30, 43),
                            _name((30, 35), "world"),
                            _type((37, 43), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_type_with_description_multi_line_string():
    assert_node_equal(
        parse(
            u'''
"""
Description
"""
# Even with comments between them
type Hello {
  world: String
}'''
        ),
        _ast.Document(
            loc=(0, 85),
            definitions=[
                _ast.ObjectTypeDefinition(
                    loc=(1, 85),
                    name=_name((60, 65), "Hello"),
                    interfaces=[],
                    directives=[],
                    description=_ast.StringValue(
                        loc=(1, 20), value="Description", block=True
                    ),
                    fields=[
                        _field(
                            (70, 83),
                            _name((70, 75), "world"),
                            _type((77, 83), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_extension():
    body = u"""
extend type Hello {
  world: String
}
"""
    assert_node_equal(
        parse(body),
        _doc(
            (0, 39),
            [
                _ast.ObjectTypeExtension(
                    loc=(1, 38),
                    name=_name((13, 18), "Hello"),
                    interfaces=[],
                    directives=[],
                    fields=[
                        _field(
                            (23, 36),
                            _name((23, 28), "world"),
                            _type((30, 36), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_extension_without_fields():
    assert_node_equal(
        parse(u"extend type Hello implements Greeting"),
        _doc(
            (0, 37),
            [
                _ast.ObjectTypeExtension(
                    loc=(0, 37),
                    name=_name((12, 17), "Hello"),
                    interfaces=[_type((29, 37), "Greeting")],
                    fields=[],
                    directives=[],
                )
            ],
        ),
    )


def test_it_parses_extension_without_fields_followed_by_extension():
    assert_node_equal(
        parse(
            """
      extend type Hello implements Greeting

      extend type Hello implements SecondGreeting
    """
        ),
        _ast.Document(
            loc=(0, 100),
            definitions=[
                _ast.ObjectTypeExtension(
                    loc=(7, 44),
                    name=_name((19, 24), "Hello"),
                    interfaces=[_type((36, 44), "Greeting")],
                    fields=[],
                    directives=[],
                ),
                _ast.ObjectTypeExtension(
                    loc=(52, 95),
                    name=_name((64, 69), "Hello"),
                    interfaces=[_type((81, 95), "SecondGreeting")],
                    fields=[],
                    directives=[],
                ),
            ],
        ),
    )


def test_extension_without_anything_throws():
    with pytest.raises(UnexpectedEOF) as exc_info:
        parse(u"extend type Hello")
    assert exc_info.value.position == 17
    assert exc_info.value.message == "Unexpected <EOF>"


def test_extension_do_not_include_descriptions_0():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(
            u"""
      "Description"
      extend type Hello {
        world: String
      }"""
        )
    assert exc_info.value.position == 27
    assert exc_info.value.message == 'Unexpected "extend"'


def test_extension_do_not_include_descriptions_1():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(
            u"""
      extend "Description" type Hello {
        world: String
      }"""
        )
    assert exc_info.value.position == 14
    assert exc_info.value.message == 'Unexpected "Description"'


def test_it_parses_simple_non_null_type():
    body = u"""
type Hello {
  world: String!
}"""
    assert_node_equal(
        parse(body),
        _ast.Document(
            loc=(0, 32),
            definitions=[
                _ast.ObjectTypeDefinition(
                    loc=(1, 32),
                    name=_name((6, 11), "Hello"),
                    interfaces=[],
                    directives=[],
                    fields=[
                        _field(
                            (16, 30),
                            _name((16, 21), "world"),
                            _ast.NonNullType(
                                loc=(23, 30), type=_type((23, 29), "String")
                            ),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_type_inheriting_interface():
    body = u"type Hello implements World { field: String }"
    assert_node_equal(
        parse(body),
        _ast.Document(
            loc=(0, 45),
            definitions=[
                _ast.ObjectTypeDefinition(
                    loc=(0, 45),
                    name=_name((5, 10), "Hello"),
                    interfaces=[_type((22, 27), "World")],
                    directives=[],
                    fields=[
                        _field(
                            (30, 43),
                            _name((30, 35), "field"),
                            _type((37, 43), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_type_inheriting_multiple_interfaces():
    body = u"type Hello implements Wo & rld { field: String }"
    assert_node_equal(
        parse(body),
        _ast.Document(
            loc=(0, 48),
            definitions=[
                _ast.ObjectTypeDefinition(
                    loc=(0, 48),
                    name=_name((5, 10), "Hello"),
                    interfaces=[_type((22, 24), "Wo"), _type((27, 30), "rld")],
                    directives=[],
                    fields=[
                        _field(
                            (33, 46),
                            _name((33, 38), "field"),
                            _type((40, 46), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_type_inheriting_multiple_interfaces_with_leading_ampersand():  # noqa: E501
    body = u"type Hello implements & Wo & rld { field: String }"
    assert_node_equal(
        parse(body),
        _ast.Document(
            loc=(0, 50),
            definitions=[
                _ast.ObjectTypeDefinition(
                    loc=(0, 50),
                    name=_name((5, 10), "Hello"),
                    interfaces=[_type((24, 26), "Wo"), _type((29, 32), "rld")],
                    directives=[],
                    fields=[
                        _field(
                            (35, 48),
                            _name((35, 40), "field"),
                            _type((42, 48), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_single_value_enum():
    body = u"enum Hello { WORLD }"
    assert_node_equal(
        parse(body),
        _doc(
            (0, 20),
            [
                _ast.EnumTypeDefinition(
                    loc=(0, 20),
                    name=_name((5, 10), "Hello"),
                    directives=[],
                    values=[
                        _ast.EnumValueDefinition(
                            loc=(13, 18), name=_name((13, 18), "WORLD")
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_double_value_enum():
    body = u"enum Hello { WO, RLD }"
    assert_node_equal(
        parse(body),
        _doc(
            (0, 22),
            [
                _ast.EnumTypeDefinition(
                    loc=(0, 22),
                    name=_name((5, 10), "Hello"),
                    directives=[],
                    values=[
                        _ast.EnumValueDefinition(
                            loc=(13, 15), name=_name((13, 15), "WO")
                        ),
                        _ast.EnumValueDefinition(
                            loc=(17, 20), name=_name((17, 20), "RLD")
                        ),
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_interface():
    body = u"""
interface Hello {
  world: String
}"""
    assert_node_equal(
        parse(body),
        _doc(
            (0, 36),
            [
                _ast.InterfaceTypeDefinition(
                    loc=(1, 36),
                    name=_name((11, 16), "Hello"),
                    directives=[],
                    fields=[
                        _field(
                            (21, 34),
                            _name((21, 26), "world"),
                            _type((28, 34), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_field_with_arg():
    body = u"""
type Hello {
  world(flag: Boolean): String
}"""
    assert_node_equal(
        parse(body),
        _doc(
            (0, 46),
            [
                _ast.ObjectTypeDefinition(
                    loc=(1, 46),
                    name=_name((6, 11), "Hello"),
                    interfaces=[],
                    directives=[],
                    fields=[
                        _field(
                            (16, 44),
                            _name((16, 21), "world"),
                            _type((38, 44), "String"),
                            [
                                _input(
                                    (22, 35),
                                    _name((22, 26), "flag"),
                                    _type((28, 35), "Boolean"),
                                )
                            ],
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_field_with_arg_with_default_value():
    body = u"""
type Hello {
  world(flag: Boolean = true): String
}"""
    assert_node_equal(
        parse(body),
        _doc(
            (0, 53),
            [
                _ast.ObjectTypeDefinition(
                    loc=(1, 53),
                    name=_name((6, 11), "Hello"),
                    interfaces=[],
                    directives=[],
                    fields=[
                        _field(
                            (16, 51),
                            _name((16, 21), "world"),
                            _type((45, 51), "String"),
                            [
                                _input(
                                    (22, 42),
                                    _name((22, 26), "flag"),
                                    _type((28, 35), "Boolean"),
                                    _ast.BooleanValue(loc=(38, 42), value=True),
                                )
                            ],
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_field_with_list_arg():
    body = u"""
type Hello {
  world(things: [String]): String
}"""
    assert_node_equal(
        parse(body),
        _doc(
            (0, 49),
            [
                _ast.ObjectTypeDefinition(
                    loc=(1, 49),
                    name=_name((6, 11), "Hello"),
                    interfaces=[],
                    directives=[],
                    fields=[
                        _field(
                            (16, 47),
                            _name((16, 21), "world"),
                            _type((41, 47), "String"),
                            [
                                _input(
                                    (22, 38),
                                    _name((22, 28), "things"),
                                    _ast.ListType(
                                        loc=(30, 38),
                                        type=_type((31, 37), "String"),
                                    ),
                                )
                            ],
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_field_with_two_args():
    body = u"""
type Hello {
  world(argOne: Boolean, argTwo: Int): String
}"""
    assert_node_equal(
        parse(body),
        _doc(
            (0, 61),
            [
                _ast.ObjectTypeDefinition(
                    loc=(1, 61),
                    name=_name((6, 11), "Hello"),
                    interfaces=[],
                    directives=[],
                    fields=[
                        _field(
                            (16, 59),
                            _name((16, 21), "world"),
                            _type((53, 59), "String"),
                            [
                                _input(
                                    (22, 37),
                                    _name((22, 28), "argOne"),
                                    _type((30, 37), "Boolean"),
                                ),
                                _input(
                                    (39, 50),
                                    _name((39, 45), "argTwo"),
                                    _type((47, 50), "Int"),
                                ),
                            ],
                        )
                    ],
                )
            ],
        ),
    )


def test_it_parses_simple_union():
    assert_node_equal(
        parse("union Hello = World"),
        _doc(
            (0, 19),
            [
                _ast.UnionTypeDefinition(
                    loc=(0, 19),
                    name=_name((6, 11), "Hello"),
                    directives=[],
                    types=[_type((14, 19), "World")],
                )
            ],
        ),
    )


def test_it_parses_union_with_two_types():
    assert_node_equal(
        parse("union Hello = Wo | Rld"),
        _doc(
            (0, 22),
            [
                _ast.UnionTypeDefinition(
                    loc=(0, 22),
                    name=_name((6, 11), "Hello"),
                    directives=[],
                    types=[_type((14, 16), "Wo"), _type((19, 22), "Rld")],
                )
            ],
        ),
    )


def test_it_parses_union_with_two_types_and_leading_pipe():
    assert_node_equal(
        parse("union Hello = | Wo | Rld"),
        _doc(
            (0, 24),
            [
                _ast.UnionTypeDefinition(
                    loc=(0, 24),
                    name=_name((6, 11), "Hello"),
                    directives=[],
                    types=[_type((16, 18), "Wo"), _type((21, 24), "Rld")],
                )
            ],
        ),
    )


def test_union_fails_with_no_types():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(u"union Hello = |")
    assert exc_info.value.position == 15
    assert exc_info.value.message == 'Expected Name but found "<EOF>"'


def test_union_fails_with_leading_douple_pipe():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(u"union Hello = || Wo | Rld")
    assert exc_info.value.position == 15
    assert exc_info.value.message == 'Expected Name but found "|"'


def test_union_fails_with_double_pipe():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(u"union Hello = Wo || Rld")
    assert exc_info.value.position == 18
    assert exc_info.value.message == 'Expected Name but found "|"'


def test_union_fails_with_trailing_pipe():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(u"union Hello = | Wo | Rld |")
    assert exc_info.value.position == 26
    assert exc_info.value.message == 'Expected Name but found "<EOF>"'


def test_it_parses_scalar():
    assert_node_equal(
        parse("scalar Hello"),
        _doc(
            (0, 12),
            [
                _ast.ScalarTypeDefinition(
                    loc=(0, 12), name=_name((7, 12), "Hello"), directives=[]
                )
            ],
        ),
    )


def test_it_parses_simple_input_object():
    body = u"""
input Hello {
  world: String
}"""
    assert_node_equal(
        parse(body),
        _doc(
            (0, 32),
            [
                _ast.InputObjectTypeDefinition(
                    loc=(1, 32),
                    name=_name((7, 12), "Hello"),
                    directives=[],
                    fields=[
                        _input(
                            (17, 30),
                            _name((17, 22), "world"),
                            _type((24, 30), "String"),
                        )
                    ],
                )
            ],
        ),
    )


def test_simple_input_object_with_args_should_fail():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(
            u"""
      input Hello {
        world(foo: Int): String
      }"""
        )

    assert exc_info.value.position == 34
    assert exc_info.value.message == 'Expected Colon but found "("'


def test_directive_with_incorrect_locations_fails():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(
            u"""
      directive @foo on FIELD | INCORRECT_LOCATION"""
        )

    assert exc_info.value.position == 33
    assert exc_info.value.message == "Unexpected Name INCORRECT_LOCATION"


class TestAllowLegacySdlEmptyFieldsOption(object):
    def test_support_type_with_empty_fields(self):
        body = u"type Hello { }"

        with pytest.raises(UnexpectedToken) as exc_info:
            parse(body)

        assert exc_info.value.position == 13
        assert exc_info.value.message == 'Expected Name but found "}"'

        assert_node_equal(
            parse(body, allow_legacy_sdl_empty_fields=True),
            _doc(
                (0, 14),
                [
                    _ast.ObjectTypeDefinition(
                        loc=(0, 14), name=_name((5, 10), "Hello")
                    )
                ],
            ),
        )


class TestAllowLegacySdlImplementsInterfacesOption(object):
    def test_it_works(self):
        body = u"type Hello implements Wo rld { field: String }"

        with pytest.raises(UnexpectedToken) as exc_info:
            parse(body)

        assert exc_info.value.position == 25
        assert exc_info.value.message == 'Unexpected "rld"'

        assert_node_equal(
            parse(body, allow_legacy_sdl_implements_interfaces=True),
            _doc(
                (0, 46),
                [
                    _ast.ObjectTypeDefinition(
                        loc=(0, 46),
                        name=_name((5, 10), "Hello"),
                        interfaces=[
                            _type((22, 24), "Wo"),
                            _type((25, 28), "rld"),
                        ],
                        fields=[
                            _field(
                                (31, 44),
                                _name((31, 36), "field"),
                                _type((38, 44), "String"),
                            )
                        ],
                    )
                ],
            ),
        )


def test_it_parses_kitchen_sink(fixture_file):
    # assert doesn't raise
    assert parse(fixture_file("schema-kitchen-sink.graphql"), no_location=True)


def test_it_does_not_parses_kitchen_sink_when_allow_type_system_is_false(
    fixture_file
):
    with pytest.raises(UnexpectedToken):
        parse(
            fixture_file("schema-kitchen-sink.graphql"),
            no_location=True,
            allow_type_system=False,
        )



# -*- coding: utf-8 -*-
""" Helpers to work with futures and the concurrent module in a way similar
to promises """

from concurrent.futures import Future

_UNDEF = object()


def is_deferred(value, cache={}):  # pylint: disable = dangerous-default-value
    t = value.__class__
    if t not in cache:
        cache[t] = callable(getattr(t, "add_done_callback", None))
    return cache[t]


def deferred(value, cls=Future):
    """ Transform a value into a ``Future`` awaitable object.

    :type value: any
    :param value: Original value

    :rtype: concurrent.futures.Future
    :returns: Value wrapped in a ``Future``

    >>> v = deferred(1, cls=DummyFuture)
    >>> v.result(), v.done()
    (1, True)
    """
    future = cls()
    future.set_result(value)
    return future


def all_(futures, cls=Future):
    """ Create a ``concurrent.futures.Future`` wrapping a list of futures.

    - The resulting future resolves only when all futures have resolved.
    - If any future rejects, the wrapping future rejects with the exception for
      that future.
    - The created future does not need an executor is considered running by
      default. As a result it cannot be cancelled, however if any of the inner
      futures is cancelled the ``concurrent.futures.CancelledError`` will be
      considered as an exception and propagated to the resulting future.

    :type futures: List[concurrent.futures.Future]
    :param futures: List of futures.

    :rtype: concurrent.futures.Future
    :returns: Single future
    """

    if not futures:
        return []

    result = cls()
    results_list = [_UNDEF] * len(futures)
    done_count = [0]
    len_ = len(futures)

    def notify():
        if done_count[0] == len_:
            result.set_result(results_list)

    def callback_for(index):
        def callback(future):

            try:
                res = future.result()
            except Exception as err:  # pylint: disable = broad-except
                result.set_exception(err)
            else:
                results_list[index] = res
                done_count[0] += 1
                notify()

        return callback

    for index, future in enumerate(futures):
        if is_deferred(future):
            future.add_done_callback(callback_for(index))
        else:
            results_list[index] = future
            done_count[0] += 1

    if done_count[0] == len_:
        return results_list
    return result


def _chain_one(source_future, map_=lambda x: x, cls=Future):
    result = cls()

    def _callback(future):
        try:
            res = map_(future.result())
        except Exception as err:  # pylint: disable = broad-except
            result.set_exception(err)
        else:
            result.set_result(res)

    source_future.add_done_callback(_callback)
    return result


def chain(leader, *funcs, **kwargs):
    """ Chain future factories together until all steps have been exhausted.

    - Final result can be either a deferred value or not, depends on the return
      type of the last step in chain. Use :func:`unwrap` to ensure the final
      result is not a future.
    - Assumes futures are independantly submitted to an executor in order to
      work without context knowledge.
    - The created future does not need an executor is considered running by
      default. As a result it cannot be cancelled, however if any of the inner
      futures is cancelled the ``concurrent.futures.CancelledError`` will be
      considered as an exception and propagated to the resulting future.

    :type leader: concurrrent.futures.Future
    :param leader: First future in the chain.
        Can be ommitted, in which case the first step will be called with
        ``None`` as input.

    :type funcs: Iterable[(any) -> any]
    :param funcs: Steps in the chain
        Each step receives the result of the previous step as input.
        Return value of a step can be either a future or a value which will be
        wrapped as a future if there is more steps to compute.

    :rtype: concurrent.futures.Future
    """
    stack = iter(funcs)
    cls = kwargs.get("cls", Future)

    def _next(value):
        if is_deferred(value):
            return unwrap(_chain_one(value, _next, cls=cls), cls=cls)
        try:
            func = next(stack)
        except StopIteration:
            return value
        else:
            return _next(func(value))

    return _next(leader)


def unwrap(source_future, cls=Future):
    """ Resolve nested futures until a non future is resolved or an
    exception is raised.

    :type source_future: concurrent.futures.Future
    :param source_future: Future to unwrap

    :rtype: any
    :returns: Unwrapped value
    """
    if not is_deferred(source_future):
        return deferred(source_future, cls=Future)

    result = cls()

    def callback(future):
        try:
            res = future.result()
        except Exception as err:  # pylint: disable = broad-except
            result.set_exception(err)
        else:
            if is_deferred(res):
                res.add_done_callback(callback)
            else:
                result.set_result(res)

    source_future.add_done_callback(callback)
    return result


def serial(steps, cls=Future):
    """ Similar to :func:`chain` but ignoring the intermediate results.

    Each step is called only after the result of the previous step has
    resolved. The resulting future rejects on the first step that rejects.

    :type steps: Iterable[() -> concurrent.futures.Future]
    :param steps:

    :rtype: concurrent.futures.Future
    :returns: Wrapped promise
    """

    def _step(original):
        return lambda _: original()

    return chain(None, *[_step(step) for step in steps], cls=cls)


def except_(
    source_future, exc_cls=(Exception,), map_=lambda x: None, cls=Future
):
    """ Except for futures.

    :type source_future: concurrent.futures.Future
    :param source_future: Future to wrap

    :type exc_cls: Union[type, Tuple[*type]]
    :param exc_cls: Exception classes to expect.
        Can be any value compatible with a standard ``except`` clause.

    :type map_: Optional[callable]
    :param map_:
        Will be passed the expected exception to generate the wrapped future's
        result.
        Default behaviour is to set the result to ``None``.
    """
    result = cls()

    def callback(future):
        try:
            res = future.result()
        except exc_cls as err:
            result.set_result(map_(err))
        except Exception as err:  # pylint: disable = broad-except
            result.set_exception(err)
        else:
            result.set_result(res)

    source_future.add_done_callback(callback)
    return result


class DummyFuture(Future):
    """ Dummy Future to match the interface in a synchronous & single
    threaded environment without the synchronisation overhead. """

    __slots__ = "_result", "_exception", "_callbacks", "_done"

    # pylint: disable = super-init-not-called
    def __init__(self):
        self._done = False
        self._result = None
        self._exception = None
        self._callbacks = []

    def _invoke_callbacks(self):
        for callback in self._callbacks:
            callback(self)

    def __repr__(self):
        if self._done:
            if self._exception:
                return "<%s at %#x state=FINISHED raised %s>" % (
                    self.__class__.__name__,
                    id(self),
                    self._exception.__class__.__name__,
                )
            else:
                return "<%s at %#x state=FINISHED returned %s>" % (
                    self.__class__.__name__,
                    id(self),
                    self._result.__class__.__name__,
                )
        return "<%s at %#x state=RUNNING>" % (self.__class__.__name__, id(self))

    def cancel(self):
        return False

    def cancelled(self):
        return False

    def running(self):
        return not self._done

    def done(self):
        return self._done

    def add_done_callback(self, fn):
        if not self._done:
            self._callbacks.append(fn)
        else:
            fn(self)

    def result(self, timeout=None):
        if not self._done:
            raise RuntimeError(
                "DummyFuture does not support blocking for results"
            )

        if self._exception:
            raise self._exception  # pylint:disable=raising-bad-type
        else:
            return self._result

    def exception(self, timeout=None):
        if not self._done:
            raise RuntimeError(
                "DummyFuture does not support blocking for results"
            )

        if self._exception:
            return self._exception
        else:
            return None

    def set_running_or_notify_cancel(self):
        pass

    def set_result(self, result):
        self._result = result
        self._done = True
        self._invoke_callbacks()

    def set_exception(self, exception):
        self._exception = exception
        self._done = True
        self._invoke_callbacks()



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import UniqueInputFieldNamesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_input_object_with_fields(schema):
    run_test(
        UniqueInputFieldNamesChecker,
        schema,
        """
    {
        field(arg: { f: true })
    }
    """,
    )


def test_same_input_object_within_two_args(schema):
    run_test(
        UniqueInputFieldNamesChecker,
        schema,
        """
    {
        field(arg1: { f: true }, arg2: { f: true })
    }
    """,
    )


def test_multiple_input_object_fields(schema):
    run_test(
        UniqueInputFieldNamesChecker,
        schema,
        """
    {
        field(arg: { f1: "value", f2: "value", f3: "value" })
    }
    """,
    )


def test_allows_for_nested_input_objects_with_similar_fields(schema):
    run_test(
        UniqueInputFieldNamesChecker,
        schema,
        """
    {
        field(arg: {
            deep: {
                deep: {
                    id: 1
                }
                id: 1
            }
            id: 1
        })
    }
    """,
    )


def test_duplicate_input_object_fields(schema):
    run_test(
        UniqueInputFieldNamesChecker,
        schema,
        """
    {
        field(arg: { f1: "value", f1: "value" })
    }
    """,
        ["There can be only one input field named f1."],
        [(41, 52)],
    )


def test_many_duplicate_input_object_fields(schema):
    run_test(
        UniqueInputFieldNamesChecker,
        schema,
        """
    {
        field(arg: { f1: "value", f1: "value", f1: "value" })
    }
    """,
        [
            "There can be only one input field named f1.",
            "There can be only one input field named f1.",
        ],
        [(41, 52), (54, 65)],
    )



# -*- coding: utf-8 -*-
"""
"""

from .._utils import DefaultOrderedDict, OrderedDict, deduplicate
from ..exc import ValidationError
from ..lang.visitor import DispatchingVisitor


class ValidationVisitor(DispatchingVisitor):
    """ Visitor class used for validating GraphQL documents.

    Subclass this to implement custom validators. Use :meth:`add_error` to
    register errors and :class:`py_gql.lang.visitor.SkipNode` to prevent
    validating child nodes when parent node is invalid.

    Args:
        schema (py_gql.schema.Schema):
            Schema to validate against (for known types, directives, etc.).
        type_info (py_gql.utilities.TypeInfoVisitor):

    Attributes:
        schema (py_gql.schema.Schema):
            Schema to validate against (for known types, directives, etc.).
        type_info (py_gql.utilities.TypeInfoVisitor):
        errors (List[py_gql.exc.ValidationError]):
    """

    def __init__(self, schema, type_info):
        super(ValidationVisitor, self).__init__()
        self.schema = schema
        self.type_info = type_info
        self.errors = []

    def add_error(self, message, nodes=None):
        """ Register an error

        Args:
            message (str): Error description
            nodes (Optional[List[py_gql.lang.ast.Node]]):
                Nodes where the error comes from
        """
        self.errors.append(ValidationError(message, nodes))


class VariablesCollector(ValidationVisitor):
    """ Custom validation visitor which tracks all variable definitions and
    usage across the document.

    This replaces ``getRecursiveVariableUsages`` and ``getVariableUsages``
    from the ref implementation and allows to work with variables without
    eagerly visitng subtrees.

    Most validation must happen in :meth:`leave_document` though.
    """

    def __init__(self, schema, type_info):
        super(VariablesCollector, self).__init__(schema, type_info)

        self._op = None
        self._op_variables = DefaultOrderedDict(OrderedDict)
        self._op_defined_variables = DefaultOrderedDict(OrderedDict)
        self._op_fragments = DefaultOrderedDict(list)
        self._fragment = None
        self._fragment_variables = DefaultOrderedDict(OrderedDict)
        self._fragment_fragments = DefaultOrderedDict(list)
        self._in_var_def = False

    def enter_operation_definition(self, node):
        self._op = node.name.value if node.name else ""

    def leave_operation_definition(self, node):
        self._op = None

    def enter_fragment_definition(self, node):
        self._fragment = node.name.value

    def leave_fragment_definition(self, node):
        self._fragment = None

    def enter_fragment_spread(self, node):
        name = node.name.value
        if self._op is not None:
            self._op_fragments[self._op].append(name)
        elif self._fragment is not None and name != self._fragment:
            self._fragment_fragments[self._fragment].append(name)

    def enter_variable_definition(self, node):
        self._in_var_def = True
        name = node.variable.name.value
        if self._op is not None:
            self._op_defined_variables[self._op][name] = node

    def leave_variable_definition(self, node):
        self._in_var_def = False

    def enter_variable(self, node):
        var = node.name.value
        input_type = self.type_info.input_type
        input_value_def = self.type_info.input_value_def
        if self._in_var_def:
            pass
        elif self._op is not None:
            self._op_variables[self._op][var] = (
                node,
                input_type,
                input_value_def,
            )
        elif self._fragment is not None:
            self._fragment_variables[self._fragment][var] = (
                node,
                input_type,
                input_value_def,
            )

    def _flatten_fragments(self):
        for parent, children in self._fragment_fragments.items():
            for child in deduplicate(children):
                for op in self._op_fragments.keys():
                    if parent in self._op_fragments[op]:
                        self._op_fragments[op].append(child)

    def leave_document(self, _):
        self._flatten_fragments()



# -*- coding: utf-8 -*-
"""
"""

from ..._string_utils import infer_suggestions, quoted_options_list
from ...exc import ScalarParsingError, UnknownEnumValue
from ...lang.visitor import SkipNode
from ...schema import (
    EnumType,
    InputObjectType,
    NonNullType,
    ScalarType,
    unwrap_type,
)
from ...schema.scalars import SPECIFIED_SCALAR_TYPES
from ..visitors import ValidationVisitor


class ValuesOfCorrectTypeChecker(ValidationVisitor):
    """ A GraphQL document is only valid if all value literals are of the type
    expected at their position. """

    # WARN: This check ignores cases where the input type is not known, which
    # should be caught by other validators.

    def _report_bad_value(self, input_type, node, extra=None):
        msg = "Expected type %s, found %s" % (input_type, node)
        if extra:
            msg += " (%s)" % extra
        self.add_error(msg, [node])

    def _check_scalar(self, node):
        input_type = self.type_info.input_type
        if not input_type:
            return

        named_type = unwrap_type(input_type)
        if not isinstance(named_type, ScalarType):
            self._report_bad_value(input_type, node)
        else:
            try:
                named_type.parse_literal(node)
            except ScalarParsingError as err:
                is_custom = named_type not in SPECIFIED_SCALAR_TYPES
                extra = str(err) if is_custom else None
                # Preserve message for custom scalar types.
                self._report_bad_value(input_type, node, extra=extra)

    enter_int_value = _check_scalar
    enter_float_value = _check_scalar
    enter_string_value = _check_scalar
    enter_boolean_value = _check_scalar

    def enter_null_value(self, node):
        input_type = self.type_info.input_type
        if input_type and isinstance(input_type, NonNullType):
            self._report_bad_value(input_type, node)

    def enter_enum_value(self, node):
        input_type = unwrap_type(self.type_info.input_type)
        if not input_type:
            return

        if not isinstance(input_type, EnumType):
            self._check_scalar(node)
        else:
            try:
                input_type.get_value(node.value)
            except UnknownEnumValue:
                self._report_bad_value(input_type, node)

    def enter_object_value(self, node):
        named_type = unwrap_type(self.type_info.input_type)
        if not isinstance(named_type, InputObjectType):
            self._check_scalar(node)
            raise SkipNode()

        input_fields = [f.name.value for f in node.fields]
        for field_def in named_type.fields:
            if field_def.required and field_def.name not in input_fields:
                self.add_error(
                    "Required field %s.%s of type %s was not provided"
                    % (named_type.name, field_def.name, field_def.type),
                    [node],
                )

    def enter_object_field(self, node):
        parent_type = unwrap_type(self.type_info.parent_input_type)
        field_type = self.type_info.input_type
        if field_type is None and isinstance(parent_type, InputObjectType):
            suggestions = infer_suggestions(
                node.name.value, [f.name for f in parent_type.fields]
            )
            if suggestions:
                self.add_error(
                    "Field %s is not defined by type %s, did you mean %s?"
                    % (
                        node.name.value,
                        parent_type,
                        quoted_options_list(suggestions),
                    ),
                    [node],
                )
            else:
                self.add_error(
                    "Field %s is not defined by type %s"
                    % (node.name.value, parent_type),
                    [node],
                )



# -*- coding: utf-8 -*-
# pylint: disable = redefined-outer-name
""" execution tests related to abstract types (Interface, Union) """
from __future__ import unicode_literals

from py_gql.schema import (
    Boolean,
    Field,
    InterfaceType,
    ListType,
    ObjectType,
    Schema,
    String,
    UnionType,
)

from ._test_utils import check_execution


class Dog(object):
    def __init__(self, name, woofs):
        self.name = name
        self.woofs = woofs


class Cat(object):
    def __init__(self, name, meows):
        self.name = name
        self.meows = meows


class Human(object):
    def __init__(self, name):
        self.name = name


class Person(object):
    def __init__(self, name, pets, friends):
        self.name = name
        self.pets = pets
        self.friends = friends


def test_ObjectType_is_type_of_for_interface_runtime_inference():
    PetType = InterfaceType("Pet", [Field("name", String)])

    DogType = ObjectType(
        "Dog",
        [Field("name", String), Field("woofs", Boolean)],
        interfaces=[PetType],
        is_type_of=Dog,
    )

    CatType = ObjectType(
        "Cat",
        [Field("name", String), Field("meows", Boolean)],
        interfaces=[PetType],
        is_type_of=Cat,
    )

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "pets",
                    ListType(PetType),
                    resolve=lambda *_: [
                        Dog("Odie", True),
                        Cat("Garfield", False),
                    ],
                )
            ],
        ),
        types=[DogType, CatType],
    )

    check_execution(
        schema,
        """{
            pets {
                name
                __typename
                ... on Dog { woofs }
                ... on Cat { meows }
            }
        }""",
        expected_data={
            "pets": [
                {"name": "Odie", "woofs": True, "__typename": "Dog"},
                {"name": "Garfield", "meows": False, "__typename": "Cat"},
            ]
        },
        expected_errors=[],
    )


def test_ObjectType_is_type_of_for_union_runtime_inference():
    DogType = ObjectType(
        "Dog", [Field("name", String), Field("woofs", Boolean)], is_type_of=Dog
    )

    CatType = ObjectType(
        "Cat", [Field("name", String), Field("meows", Boolean)], is_type_of=Cat
    )

    PetType = UnionType("Pet", [DogType, CatType])

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "pets",
                    ListType(PetType),
                    resolve=lambda *_: [
                        Dog("Odie", True),
                        Cat("Garfield", False),
                    ],
                )
            ],
        ),
        types=[DogType, CatType],
    )

    check_execution(
        schema,
        """{
            pets {
                __typename
                ... on Dog { woofs, name }
                ... on Cat { meows, name }
            }
        }""",
        expected_data={
            "pets": [
                {"name": "Odie", "woofs": True, "__typename": "Dog"},
                {"name": "Garfield", "meows": False, "__typename": "Cat"},
            ]
        },
        expected_errors=[],
    )


def test_type_resolution_on_interface_yields_useful_error():
    """ Different from ref implementation -> this should never happen
    so we crash """

    def _resolve_pet_type(value, **_):
        return {Dog: DogType, Cat: CatType, Human: HumanType}.get(
            type(value), None
        )

    PetType = InterfaceType(
        "Pet", [Field("name", String)], resolve_type=_resolve_pet_type
    )

    HumanType = ObjectType("Human", [Field("name", String)])

    DogType = ObjectType(
        "Dog",
        [Field("name", String), Field("woofs", Boolean)],
        interfaces=[PetType],
    )

    CatType = ObjectType(
        "Cat",
        [Field("name", String), Field("meows", Boolean)],
        interfaces=[PetType],
    )

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "pets",
                    ListType(PetType),
                    resolve=lambda *_: [
                        Dog("Odie", True),
                        Cat("Garfield", False),
                        Human("Jon"),
                    ],
                )
            ],
        ),
        types=[DogType, CatType],
    )

    check_execution(
        schema,
        """{
        pets {
            name
            __typename
            ... on Dog { woofs }
            ... on Cat { meows }
        }
        }""",
        expected_exc=RuntimeError,
        expected_msg=(
            'Runtime ObjectType "Human" is not a possible type for field '
            '"pets[2]" of type "Pet".'
        ),
    )


def test_type_resolution_on_union_yields_useful_error():
    """ Different from ref implementation -> this should never happen
    so we crash """

    def _resolve_pet_type(value, **_):
        return {Dog: DogType, Cat: CatType, Human: HumanType}.get(
            type(value), None
        )

    PetType = InterfaceType(
        "Pet", [Field("name", String)], resolve_type=_resolve_pet_type
    )

    HumanType = ObjectType("Human", [Field("name", String)])

    DogType = ObjectType(
        "Dog", [Field("name", String), Field("woofs", Boolean)]
    )

    CatType = ObjectType(
        "Cat", [Field("name", String), Field("meows", Boolean)]
    )

    PetType = UnionType(
        "Pet", [DogType, CatType], resolve_type=_resolve_pet_type
    )

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "pets",
                    ListType(PetType),
                    resolve=lambda *_: [
                        Dog("Odie", True),
                        Cat("Garfield", False),
                        Human("Jon"),
                    ],
                )
            ],
        ),
        types=[DogType, CatType],
    )

    check_execution(
        schema,
        """{
        pets {
            __typename
            ... on Dog { woofs, name }
            ... on Cat { meows, name }
        }
        }""",
        expected_exc=RuntimeError,
        expected_msg=(
            'Runtime ObjectType "Human" is not a possible type for field '
            '"pets[2]" of type "Pet".'
        ),
    )


def test_type_resolution_supports_strings():
    def _resolve_pet_type(value, **_):
        return type(value).__name__

    PetType = InterfaceType(
        "Pet", [Field("name", String)], resolve_type=_resolve_pet_type
    )

    DogType = ObjectType(
        "Dog",
        [Field("name", String), Field("woofs", Boolean)],
        interfaces=[PetType],
    )

    CatType = ObjectType(
        "Cat",
        [Field("name", String), Field("meows", Boolean)],
        interfaces=[PetType],
    )

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "pets",
                    ListType(PetType),
                    resolve=lambda *_: [
                        Dog("Odie", True),
                        Cat("Garfield", False),
                    ],
                )
            ],
        ),
        types=[DogType, CatType],
    )

    check_execution(
        schema,
        """{
        pets {
            name
            __typename
            ... on Dog { woofs }
            ... on Cat { meows }
        }
        }""",
        expected_errors=[],
    )


def test_type_resolution_supports_object_attribute():
    PetType = InterfaceType("Pet", [Field("name", String)])

    DogType = ObjectType(
        "Dog",
        [Field("name", String), Field("woofs", Boolean)],
        interfaces=[PetType],
    )

    CatType = ObjectType(
        "Cat",
        [Field("name", String), Field("meows", Boolean)],
        interfaces=[PetType],
    )

    class Dog(object):
        __typename__ = DogType

        def __init__(self, name, woofs):
            self.name = name
            self.woofs = woofs

    class Cat(object):
        __typename__ = "Cat"

        def __init__(self, name, meows):
            self.name = name
            self.meows = meows

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "pets",
                    ListType(PetType),
                    resolve=lambda *_: [
                        Dog("Odie", True),
                        Cat("Garfield", False),
                    ],
                )
            ],
        ),
        types=[DogType, CatType],
    )

    check_execution(
        schema,
        """{
        pets {
            name
            __typename
            ... on Dog { woofs }
            ... on Cat { meows }
        }
        }""",
        expected_errors=[],
    )


NamedType = InterfaceType("Named", [Field("name", String)])

DogType = ObjectType(
    "Dog",
    [Field("name", String), Field("woofs", Boolean)],
    interfaces=[NamedType],
    is_type_of=Dog,
)


CatType = ObjectType(
    "Cat",
    [Field("name", String), Field("meows", Boolean)],
    interfaces=[NamedType],
    is_type_of=Cat,
)


def _resolve_pet_type(value, *a, **kw):
    if isinstance(value, Dog):
        return DogType
    elif isinstance(value, Cat):
        return CatType


PetType = UnionType("Pet", [DogType, CatType], resolve_type=_resolve_pet_type)

PersonType = ObjectType(
    "Person",
    [
        Field("name", String),
        Field("pets", ListType(PetType)),
        Field("friends", lambda: ListType(NamedType)),
    ],
    interfaces=[NamedType],
    is_type_of=Person,
)

_SCHEMA = Schema(PersonType)

_GARFIELD = Cat("Garfield", False)
_ODIE = Dog("Odie", True)
_LIZ = Person("Liz", None, None)
_JOHN = Person("John", [_GARFIELD, _ODIE], [_LIZ, _ODIE])


def test_it_can_introspect_on_union_and_intersection_types():
    check_execution(
        _SCHEMA,
        """
    {
        Named: __type(name: "Named") {
            kind
            name
            fields { name }
            interfaces { name }
            possibleTypes { name }
            enumValues { name }
            inputFields { name }
        }
        Pet: __type(name: "Pet") {
            kind
            name
            fields { name }
            interfaces { name }
            possibleTypes { name }
            enumValues { name }
            inputFields { name }
        }
    }
    """,
        {
            "Named": {
                "enumValues": None,
                "fields": [{"name": "name"}],
                "inputFields": None,
                "interfaces": None,
                "kind": "INTERFACE",
                "name": "Named",
                "possibleTypes": [
                    {"name": "Cat"},
                    {"name": "Dog"},
                    {"name": "Person"},
                ],
            },
            "Pet": {
                "enumValues": None,
                "fields": None,
                "inputFields": None,
                "interfaces": None,
                "kind": "UNION",
                "name": "Pet",
                "possibleTypes": [{"name": "Cat"}, {"name": "Dog"}],
            },
        },
        [],
    )


def test_it_executes_union_types():
    # This is an *invalid* query, but it should be an *executable* query.
    check_execution(
        _SCHEMA,
        """
    {
        __typename
        name
        pets {
            __typename
            name
            woofs
            meows
        }
    }
    """,
        {
            "__typename": "Person",
            "name": "John",
            "pets": [
                {"__typename": "Cat", "meows": False, "name": "Garfield"},
                {"__typename": "Dog", "name": "Odie", "woofs": True},
            ],
        },
        [],
        initial_value=_JOHN,
    )


def test_it_executes_union_types_using_inline_fragments():
    # This is the valid version of the query in the above test.
    check_execution(
        _SCHEMA,
        """
    {
        __typename
        name
        pets {
            __typename
            ... on Dog { name woofs }
            ... on Cat { name meows }
        }
    }
    """,
        {
            "__typename": "Person",
            "name": "John",
            "pets": [
                {"__typename": "Cat", "meows": False, "name": "Garfield"},
                {"__typename": "Dog", "name": "Odie", "woofs": True},
            ],
        },
        [],
        initial_value=_JOHN,
    )


def test_it_executes_interface_types():
    # This is an *invalid* query, but it should be an *executable* query.
    check_execution(
        _SCHEMA,
        """
    {
        __typename
        name
        friends {
            __typename
            name
            woofs
            meows
        }
    }
    """,
        {
            "__typename": "Person",
            "friends": [
                {"__typename": "Person", "name": "Liz"},
                {"__typename": "Dog", "name": "Odie", "woofs": True},
            ],
            "name": "John",
        },
        [],
        initial_value=_JOHN,
    )


def test_it_executes_interface_types_using_inline_fragments():
    # This is the valid version of the query in the above test.
    check_execution(
        _SCHEMA,
        """
    {
        __typename
        name
        friends {
            __typename
            name
            ... on Dog { woofs }
            ... on Cat { meows }
        }
    }
    """,
        {
            "__typename": "Person",
            "friends": [
                {"__typename": "Person", "name": "Liz"},
                {"__typename": "Dog", "name": "Odie", "woofs": True},
            ],
            "name": "John",
        },
        [],
        initial_value=_JOHN,
    )


def test_it_allows_fragment_conditions_to_be_abstract_types():
    check_execution(
        _SCHEMA,
        """
    {
        __typename
        name
        pets { ...PetFields }
        friends { ...FriendFields }
    }

    fragment PetFields on Pet {
        __typename
        ... on Dog { name woofs }
        ... on Cat { name meows }
    }

    fragment FriendFields on Named {
        __typename
        name
        ... on Dog { woofs }
        ... on Cat { meows }
    }
    """,
        {
            "__typename": "Person",
            "friends": [
                {"__typename": "Person", "name": "Liz"},
                {"__typename": "Dog", "name": "Odie", "woofs": True},
            ],
            "name": "John",
            "pets": [
                {"__typename": "Cat", "meows": False, "name": "Garfield"},
                {"__typename": "Dog", "name": "Odie", "woofs": True},
            ],
        },
        [],
        initial_value=_JOHN,
    )



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import ScalarLeafsChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_valid_scalar_selection(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    fragment scalarSelection on Dog {
        barks
    }""",
    )


def test_object_type_missing_selection(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    query directQueryOnObjectWithoutSubFields {
        human
    }""",
        ['Field "human" of type "Human" must have a subselection'],
    )


def test_interface_type_missing_selection(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    {
        human { pets }
    }""",
        ['Field "pets" of type "[Pet]" must have a subselection'],
    )


def test_valid_scalar_selection_with_args(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    fragment scalarSelectionWithArgs on Dog {
        doesKnowCommand(dogCommand: SIT)
    }""",
    )


def test_scalar_selection_not_allowed_on_boolean(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    fragment scalarSelectionsNotAllowedOnBoolean on Dog {
        barks { sinceWhen }
    }""",
        [
            'Field "barks" cannot have a selection as type "Boolean" '
            "has no fields"
        ],
    )


def test_scalar_selection_not_allowed_on_enum(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    fragment scalarSelectionsNotAllowedOnEnum on Cat {
        furColor { inHexdec }
    }""",
        [
            'Field "furColor" cannot have a selection as type "FurColor" '
            "has no fields"
        ],
    )


def test_scalar_selection_not_allowed_with_args(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    fragment scalarSelectionsNotAllowedWithArgs on Dog {
        doesKnowCommand(dogCommand: SIT) { sinceWhen }
    }""",
        [
            'Field "doesKnowCommand" cannot have a selection as type "Boolean" '
            "has no fields"
        ],
    )


def test_scalar_selection_not_allowed_with_directives(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    fragment scalarSelectionsNotAllowedWithDirectives on Dog {
        name @include(if: true) { isAlsoHumanName }
    }""",
        [
            'Field "name" cannot have a selection as type "String" '
            "has no fields"
        ],
    )


def test_scalar_selection_not_allowed_with_directives_and_args(schema):
    run_test(
        ScalarLeafsChecker,
        schema,
        """
    fragment scalarSelectionsNotAllowedWithDirectivesAndArgs on Dog {
        doesKnowCommand(dogCommand: SIT) @include(if: true) { sinceWhen }
    }""",
        [
            'Field "doesKnowCommand" cannot have a selection as type "Boolean" '
            "has no fields"
        ],
    )



# -*- coding: utf-8 -*-
""" Test the main async/await entry point """

import asyncio
import time
from concurrent.futures import CancelledError

import pytest

from py_gql.asyncio import AsyncIOExecutor, graphql
from py_gql.exc import SchemaError
from py_gql.execution import _concurrency
from py_gql.execution.executors import ThreadPoolExecutor, SyncExecutor
from py_gql.schema import Argument, Field, ObjectType, Schema, String


@pytest.mark.asyncio
async def test_it_correctly_identifies_r2_d2_as_the_hero_of_the_star_wars_saga(
    starwars_schema
):
    result = await graphql(
        starwars_schema,
        """
        query HeroNameQuery {
            hero {
            name
            }
        }
        """,
    )
    assert result.response() == {"data": {"hero": {"name": "R2-D2"}}}


@pytest.mark.asyncio
async def test_correct_response_on_syntax_error_1(starwars_schema):
    with AsyncIOExecutor() as executor:
        result = await graphql(starwars_schema, "", {}, executor=executor)

    assert result.response() == {
        "errors": [
            {
                "message": "Unexpected <EOF> (1:1):\n  1:\n    ^",
                "locations": [{"columne": 1, "line": 1}],
            }
        ]
    }


@pytest.mark.asyncio
async def test_correct_response_on_syntax_error_2(starwars_schema):
    query = """
    query HeroNameQuery {{
        hero {
           name
        }
    }
    """

    with AsyncIOExecutor() as executor:
        result = await graphql(starwars_schema, query, {}, executor=executor)

    assert result.response() == {
        "errors": [
            {
                "message": """Expected Name but found "{" (2:26):
  1:
  2:    query HeroNameQuery {{
                             ^
  3:        hero {
  4:           name""",
                "locations": [{"columne": 26, "line": 2}],
            }
        ]
    }


@pytest.mark.asyncio
async def test_correct_response_on_validation_errors(starwars_schema):
    query = """
    query HeroNameAndFriendsQuery($hero: Droid) {
        hero {
            id
            foo
            friends {
                name
            }
        }
    }

    fragment hero on Character {
        id
        friends { name }
    }
    """

    with AsyncIOExecutor() as executor:
        result = await graphql(starwars_schema, query, {}, executor=executor)

    assert result.response() == {
        "errors": [
            {
                "locations": [{"column": 35, "line": 2}],
                "message": 'Variable "$hero" must be input type',
            },
            {
                "locations": [{"column": 13, "line": 5}],
                "message": 'Cannot query field "foo" on type "Character"',
            },
            {"message": 'Unused fragment(s) "hero"'},
            {
                "locations": [{"column": 35, "line": 2}],
                "message": 'Unused variable "$hero"',
            },
        ]
    }


@pytest.mark.asyncio
async def test_correct_response_on_argument_validation_error(starwars_schema):
    query = """
    query HeroNameQuery {
        luke: human {
            name
        }
    }
    """
    with AsyncIOExecutor() as executor:
        result = await graphql(starwars_schema, query, {}, executor=executor)

    assert result.response() == {
        "errors": [
            {
                "message": (
                    'Field "human" argument "id" of type String! '
                    "is required but not provided"
                ),
                "locations": [{"line": 3, "column": 9}],
            }
        ]
    }


@pytest.mark.asyncio
async def test_correct_response_on_execution_error(starwars_schema):
    query = """
    query HeroNameAndFriendsQuery {
        hero {
            id
            friends {
                name
            }
        }
    }

    query HeroNameQuery {
        hero {
           name
        }
    }
    """

    with AsyncIOExecutor() as executor:
        result = await graphql(starwars_schema, query, {}, executor=executor)

    assert result.response() == {
        "errors": [
            {
                "message": "Operation name is required when document contains "
                "multiple operation definitions"
            }
        ],
        "data": None,
    }


@pytest.mark.asyncio
async def test_correct_response_on_execution_error_2(starwars_schema):
    query = """
    query HeroNameAndFriendsQuery {
        hero {
            id
            friends {
                name
            }
        }
    }

    query HeroNameQuery {
        hero {
           name
        }
    }
    """

    with AsyncIOExecutor() as executor:
        result = await graphql(
            starwars_schema, query, {}, operation_name="Foo", executor=executor
        )

    assert result.response() == {
        "errors": [{"message": 'No operation "Foo" found in document'}],
        "data": None,
    }


@pytest.mark.asyncio
async def test_correct_response_on_execution_error_3(starwars_schema):
    query = """
    mutation  {
        hero {
            id
            friends {
                name
            }
        }
    }
    """
    with AsyncIOExecutor() as executor:
        result = await graphql(starwars_schema, query, {}, executor=executor)

    assert result.response() == {
        "errors": [{"message": "Schema doesn't support mutation operation"}],
        "data": None,
    }


@pytest.mark.asyncio
async def test_correct_response_on_variables_error(starwars_schema):
    query = """
    query ($episode: Episode!, $human: String!) {
        hero(episode: $episode) {
            name
        }
        human(id: $human) {
            name
        }
    }
    """

    with AsyncIOExecutor() as executor:
        result = await graphql(
            starwars_schema, query, {"episode": 42, "id": 42}, executor=executor
        )

    assert result.response() == {
        "errors": [
            {
                "message": (
                    'Variable "$episode" got invalid value 42 '
                    "(Expected type Episode)"
                ),
                "locations": [{"line": 2, "column": 12}],
            },
            {
                "message": (
                    'Variable "$human" of required type "String!" '
                    "was not provided."
                ),
                "locations": [{"line": 2, "column": 32}],
            },
        ],
        "data": None,
    }


@pytest.mark.asyncio
async def test_correct_response_on_resolver_error(starwars_schema):
    query = """
    query HeroNameQuery {
        mainHero: hero {
            name
            story: secretBackstory
        }
    }
    """
    with AsyncIOExecutor() as executor:
        result = await graphql(starwars_schema, query, {}, executor=executor)

    assert result.response() == {
        "errors": [
            {
                "message": "secretBackstory is secret.",
                "locations": [{"line": 5, "column": 13}],
                "path": ["mainHero", "story"],
                "extensions": {"code": 42},
            }
        ],
        "data": {"mainHero": {"name": "R2-D2", "story": None}},
    }


@pytest.mark.asyncio
async def test_raises_if_invalid_schema_is_provided():
    with pytest.raises(SchemaError) as exc_info:
        await graphql(Schema(String), "{ field }", {})
    assert str(exc_info.value) == 'Query must be ObjectType but got "String"'


@pytest.mark.asyncio
async def test_async_resolvers():
    async def resolve_x(root, args, ctx, info):
        await asyncio.sleep(0.25)
        return root.get("x", None)

    obj = ObjectType("Object", [Field("x", String, resolve=resolve_x)])
    schema = Schema(obj)

    result = await graphql(schema, "{ x }", initial_value={"x": "foo"})

    assert result.response() == {"data": {"x": "foo"}}


async def _async_func(x):
    await asyncio.sleep(.1)
    return x


async def _non_async_func(x):
    return x


class TestAsyncIOExecutor(object):
    @pytest.mark.asyncio
    async def test_submit_returns_a_future(self):
        f = AsyncIOExecutor().submit(_async_func, 1)
        assert _concurrency.is_deferred(f)
        r = await f
        assert r == 1

    @pytest.mark.asyncio
    async def test_submit_accepts_non_async_functions(self):
        f = AsyncIOExecutor().submit(_non_async_func, 1)
        assert _concurrency.is_deferred(f)
        r = await f
        assert r == 1

    @pytest.mark.asyncio
    async def test_shutdowns_cancels_remaining_tasks(self):
        with AsyncIOExecutor() as e:
            f = e.submit(_async_func, 1)

        with pytest.raises(CancelledError):
            await f

    @pytest.mark.asyncio
    async def test_shutdowns_leaves_done_tasks_alone(self):
        with AsyncIOExecutor() as e:
            f = e.submit(_async_func, 1)
            assert await f == 1
        assert await f == 1

    @pytest.mark.asyncio
    async def test_rejects_futures_after_shutdown(self):
        with AsyncIOExecutor() as e:
            f = e.submit(_async_func, 1)
            await f

        with pytest.raises(RuntimeError):
            e.submit(_async_func, 1)


@pytest.mark.asyncio
async def test_it_runs_coroutines_concurrently():

    log = []

    async def resolve_and_log(root, args, *_):
        log.append(("in", args.get("id", None)))
        await asyncio.sleep(.1)
        log.append(("out", args.get("id", None)))
        return args.get("id")

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "foo",
                    String,
                    [Argument("id", String)],
                    resolve=resolve_and_log,
                )
            ],
        )
    )

    res = await graphql(schema, '{ one: foo(id: "1"), two: foo(id: "2") }')

    assert res.response() == {"data": {"one": "1", "two": "2"}}
    assert log[:2] == [("in", "1"), ("in", "2")]  # This part must be in order
    # The last 2 can be misordered depending on scheduling
    assert ("out", "1") in log
    assert ("out", "2") in log


@pytest.mark.asyncio
async def test_with_ThreadPoolExecutor():
    log = []

    def resolve_and_log(root, args, *_):
        log.append(("in", args.get("id", None)))
        time.sleep(0.1)
        log.append(("out", args.get("id", None)))
        return args.get("id")

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "foo",
                    String,
                    [Argument("id", String)],
                    resolve=resolve_and_log,
                )
            ],
        )
    )

    with ThreadPoolExecutor() as executor:
        res = await graphql(
            schema,
            '{ one: foo(id: "1"), two: foo(id: "2") }',
            executor=executor,
        )

    assert res.response() == {"data": {"one": "1", "two": "2"}}
    assert log[:2] == [("in", "1"), ("in", "2")]  # This part must be in order
    # The last 2 can be misordered depending on scheduling
    assert ("out", "1") in log
    assert ("out", "2") in log


# WARN: This is just testing that it works, you shouldn't do that in real life.
@pytest.mark.asyncio
async def test_with_SyncExecutor():
    log = []

    def resolve_and_log(root, args, *_):
        log.append(("in", args.get("id", None)))
        time.sleep(0.1)
        log.append(("out", args.get("id", None)))
        return args.get("id")

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "foo",
                    String,
                    [Argument("id", String)],
                    resolve=resolve_and_log,
                )
            ],
        )
    )

    with SyncExecutor() as executor:
        res = await graphql(
            schema,
            '{ one: foo(id: "1"), two: foo(id: "2") }',
            executor=executor,
        )

    assert res.response() == {"data": {"one": "1", "two": "2"}}
    # These are always in order due to blocking behaviour
    assert log == [("in", "1"), ("out", "1"), ("in", "2"), ("out", "2")]



# -*- coding: utf-8 -*-
"""
This defines a basic set of data and the Star Wars Schema used for testing
and examples. It describes the major characters in the original
Star Wars trilogy.

NOTE: This may contain spoilers for the original Star Wars trilogy.

NOTE: The data is hard coded for the sake of the demo, but you could imagine
fetching this data from a backend service like SWApi rather than from hardcoded
JSON objects in a more complex demo.
"""

from py_gql.exc import ResolverError
from py_gql.schema import (
    Argument,
    EnumType,
    EnumValue,
    Field,
    InterfaceType,
    ListType,
    NonNullType,
    ObjectType,
    Schema,
    String,
)

luke = {
    "type": "Human",
    "id": "1000",
    "name": "Luke Skywalker",
    "friends": ["1002", "1003", "2000", "2001"],
    "appearsIn": [4, 5, 6],
    "homePlanet": "Tatooine",
}

vader = {
    "type": "Human",
    "id": "1001",
    "name": "Darth Vader",
    "friends": ["1004"],
    "appearsIn": [4, 5, 6],
    "homePlanet": "Tatooine",
}

han = {
    "type": "Human",
    "id": "1002",
    "name": "Han Solo",
    "friends": ["1000", "1003", "2001"],
    "appearsIn": [4, 5, 6],
}

leia = {
    "type": "Human",
    "id": "1003",
    "name": "Leia Organa",
    "friends": ["1000", "1002", "2000", "2001"],
    "appearsIn": [4, 5, 6],
    "homePlanet": "Alderaan",
}

tarkin = {
    "type": "Human",
    "id": "1004",
    "name": "Wilhuff Tarkin",
    "friends": ["1001"],
    "appearsIn": [4],
}

human_data = {
    "1000": luke,
    "1001": vader,
    "1002": han,
    "1003": leia,
    "1004": tarkin,
}

threepio = {
    "type": "Droid",
    "id": "2000",
    "name": "C-3PO",
    "friends": ["1000", "1002", "1003", "2001"],
    "appearsIn": [4, 5, 6],
    "primaryFunction": "Protocol",
}

artoo = {
    "type": "Droid",
    "id": "2001",
    "name": "R2-D2",
    "friends": ["1000", "1002", "1003"],
    "appearsIn": [4, 5, 6],
    "primaryFunction": "Astromech",
}

droid_data = {"2000": threepio, "2001": artoo}


def get_character(id_):
    return get_human(id_) or get_droid(id_)


def get_friends(character):
    return [get_character(f) for f in character["friends"]]


def get_hero(episode):
    if episode == 5:
        return luke
    return artoo


def get_human(id_):
    return human_data.get(id_)


def get_droid(id_):
    return droid_data.get(id_)


Episode = EnumType(
    "Episode",
    [
        EnumValue("NEWHOPE", 4, description="Released in 1977."),
        EnumValue("EMPIRE", 5, description="Released in 1980."),
        EnumValue("JEDI", 6, description="Released in 1983."),
    ],
    description="One of the films in the Star Wars Trilogy",
)


def resolve_character_type(character, *r, **kw):
    return {"Human": Human, "Droid": Droid}[character["type"]]


Character = InterfaceType(
    "Character",
    [
        Field(
            "id", NonNullType(String), description="The id of the character."
        ),
        Field("name", String, description="The name of the character."),
        Field(
            "friends",
            ListType(lambda: Character),
            description=(
                "The friends of the character, or an empty list if they have "
                "none."
            ),
        ),
        Field(
            "appearsIn",
            ListType(Episode),
            description="Which movies they appear in.",
        ),
        Field(
            "secretBackstory",
            String,
            description="All secrets about their past.",
        ),
    ],
    description="A character in the Star Wars Trilogy",
    resolve_type=resolve_character_type,
)


def resolve_secret_backstory(*a, **kw):
    raise ResolverError("secretBackstory is secret.", extensions=[("code", 42)])


Human = ObjectType(
    "Human",
    [
        Field("id", NonNullType(String), description="The id of the human."),
        Field("name", String, description="The name of the human."),
        Field(
            "friends",
            ListType(lambda: Character),
            description=(
                "The friends of the human, or an empty list if they have "
                "none."
            ),
            resolve=lambda human, *r, **k: get_friends(human),
        ),
        Field(
            "appearsIn",
            ListType(Episode),
            description="Which movies they appear in.",
        ),
        Field(
            "secretBackstory",
            String,
            description=(
                "Where are they from and how they came to be who they are."
            ),
            resolve=resolve_secret_backstory,
        ),
        Field(
            "homePlanet",
            String,
            description="The home planet of the human, or null if unknown.",
        ),
    ],
    description="A humanoid creature in the Star Wars universe.",
    interfaces=[Character],
)

Droid = ObjectType(
    "Droid",
    [
        Field("id", NonNullType(String), description="The id of the droid."),
        Field("name", String, description="The name of the droid."),
        Field(
            "friends",
            ListType(lambda: Character),
            description=(
                "The friends of the droid, or an empty list if they have "
                "none."
            ),
            resolve=lambda droid, *r, **k: get_friends(droid),
        ),
        Field(
            "appearsIn",
            ListType(Episode),
            description="Which movies they appear in.",
        ),
        Field(
            "secretBackstory",
            String,
            description=(
                "Where are they from and how they came to be who they are."
            ),
            resolve=resolve_secret_backstory,
        ),
        Field(
            "primaryFunction",
            String,
            description="The primary function of the droid.",
        ),
    ],
    description="A mechanical creature in the Star Wars universe.",
    interfaces=[Character],
)


Query = ObjectType(
    "Query",
    [
        Field(
            "hero",
            Character,
            [
                Argument(
                    "episode",
                    Episode,
                    description=(
                        "If omitted, returns the hero of the whole saga. If "
                        "provided, returns the hero of that particular episode."
                    ),
                )
            ],
            resolve=lambda _, args, *r: get_hero(args.get("episode")),
        ),
        Field(
            "human",
            Human,
            [
                Argument(
                    "id", NonNullType(String), description="Id of the human"
                )
            ],
            resolve=lambda _, args, *r: get_human(args.get("id")),
        ),
        Field(
            "droid",
            Droid,
            [
                Argument(
                    "id", NonNullType(String), description="Id of the droid"
                )
            ],
            resolve=lambda _, args, *r: get_droid(args.get("id")),
        ),
    ],
)

StarWarsSchema = Schema(Query, types=[Human, Droid])



# -*- coding: utf-8 -*-
""" Implement schema directives.
This is based of ther way Apollo graphql handles them which is in essence an
extension of the Visitor concept.
"""

from .. import types as _types
from ..._utils import flatten
from ...exc import SDLError
from ...utilities import coerce_argument_values
from ..directives import SPECIFIED_DIRECTIVES
from ..scalars import SPECIFIED_SCALAR_TYPES
from ..schema import Schema

_SPECIFIED_DIRECTIVE_NAMES = frozenset(d.name for d in SPECIFIED_DIRECTIVES)


def visit_schema(visitor, schema):
    """ Traverse a schema based on a visitor instance.

    Args:
        visitor (SchemaVisitor): Visitor to use
        schema (py_gql.schema.Schema): Schema to visit

    Returns:
        py_gql.schema.Schema:
    """
    visitor.visit_schema(schema)

    updated_types = {}
    for type_name, old_type in schema.types.items():
        if type_name.startswith("__"):
            continue
        new_type = _visit(visitor, old_type)
        if new_type is not None and new_type is not old_type:
            updated_types[type_name] = new_type

    if updated_types:
        for k, v in updated_types.items():
            schema.type_map[k] = v
        schema._rebuild_caches()

    return schema


def _visit_and_filter(visitor, iterator):
    # type: (Visitor, Iterator[T]) -> List[T]
    visited = (_visit(visitor, entry) for entry in iterator)
    return [value for value in visited if value is not None]


def _visit(visitor, definition):
    # type: (Visitor, T) -> Optional[T]
    if isinstance(definition, Schema):
        return visit_schema(visitor, definition)
    elif isinstance(definition, _types.ObjectType):
        return _visit_object(visitor, definition)
    elif isinstance(definition, _types.InterfaceType):
        return _visit_interface(visitor, definition)
    elif isinstance(definition, _types.InputObjectType):
        return _visit_input_object(visitor, definition)
    elif isinstance(definition, _types.ScalarType):
        if definition not in SPECIFIED_SCALAR_TYPES:
            return visitor.visit_scalar(definition)
        return definition
    elif isinstance(definition, _types.UnionType):
        return visitor.visit_union(definition)
    elif isinstance(definition, _types.EnumType):
        return _visit_enum(visitor, definition)
    elif isinstance(definition, _types.Field):
        return _visit_field(visitor, definition)
    elif isinstance(definition, _types.Argument):
        return visitor.visit_argument(definition)
    elif isinstance(definition, _types.InputField):
        return visitor.visit_input_field(definition)
    elif isinstance(definition, _types.EnumValue):
        return visitor.visit_enum_value(definition)

    raise TypeError(type(definition))


def _visit_object(visitor, object_type):
    # type: (Visitor, _types.ObjectType) -> Optional[_types.ObjectType]
    new_type = visitor.visit_object(object_type)
    if new_type is not None:
        new_type.fields = _visit_and_filter(visitor, new_type.fields)
    return new_type


def _visit_interface(visitor, iface):
    # type: (Visitor, _types.InterfaceType) -> Optional[_types.InterfaceType]
    new_type = visitor.visit_interface(iface)
    if new_type is not None:
        new_type.fields = _visit_and_filter(visitor, new_type.fields)
    return new_type


def _visit_field(visitor, field):
    # type: (Visitor, _types.Field) -> Optional[_types.Field]
    new_field = visitor.visit_field(field)
    if new_field is not None:
        if new_field.args:
            new_field.args = _visit_and_filter(visitor, new_field.args)
    return new_field


def _visit_input_object(visitor, input_object):
    # type: (Visitor, _types.InputObjectType) -> Optional[_types.InputObjectType]
    new_type = visitor.visit_input_object(input_object)
    if new_type is not None:
        new_type.fields = _visit_and_filter(visitor, new_type.fields)
    return new_type


def _visit_enum(visitor, enum):
    # type: (Visitor, _types.EnumType) -> Optional[_types.EnumType]
    new_type = visitor.visit_enum(enum)
    return _types.EnumType(
        name=new_type.name,
        description=new_type.description,
        values=_visit_and_filter(visitor, new_type.values),
    )


class SchemaVisitor(object):
    """ Base class to encode schema traversal and inline modifications.

    Subclass and override the ``visit_*`` methods to implement custom behaviour.
    """

    def visit(self, definition):
        return _visit(self, definition)

    def default(self, definition):
        return definition

    # Override these methods to actually do anything to the schema
    visit_schema = lambda self, definition: self.default(definition)
    visit_scalar = lambda self, definition: self.default(definition)
    visit_object = lambda self, definition: self.default(definition)
    visit_field = lambda self, definition: self.default(definition)
    visit_argument = lambda self, definition: self.default(definition)
    visit_interface = lambda self, definition: self.default(definition)
    visit_union = lambda self, definition: self.default(definition)
    visit_enum = lambda self, definition: self.default(definition)
    visit_enum_value = lambda self, definition: self.default(definition)
    visit_input_object = lambda self, definition: self.default(definition)
    visit_input_field = lambda self, definition: self.default(definition)


_CLS_TO_LOC = {
    Schema: "SCHEMA",
    _types.ScalarType: "SCALAR",
    _types.ObjectType: "OBJECT",
    _types.Field: "FIELD_DEFINITION",
    _types.Argument: "ARGUMENT_DEFINITION",
    _types.InterfaceType: "INTERFACE",
    _types.UnionType: "UNION",
    _types.EnumType: "ENUM",
    _types.EnumValue: "ENUM_VALUE",
    _types.InputObjectType: "INPUT_OBJECT",
    _types.InputField: "INPUT_FIELD_DEFINITION",
}


def _find_directives(definition):
    node = getattr(definition, "node", None)
    if node:
        return definition.node.directives or []
    nodes = getattr(definition, "nodes", [])
    return list(flatten(node.directives or [] for node in nodes if node))


class HealSchemaVisitor(SchemaVisitor):
    """ Ensure internal representation of types match the ones in the
    schema's top level type map.

    This useful after modifying a schema inline or using a
    :class:`SchemaVisitor` instance where a type may have swapped out but not
    all references (e.g. arguments, fields, union, etc.) were. """

    def __init__(self, schema):
        self.schema = schema

    def _healed(self, original):
        if isinstance(original, (_types.NonNullType, _types.ListType)):
            return type(original)(self._healed(original.type))
        return self.schema.get_type(original.name, original)

    def visit_schema(self, schema):
        if schema.query_type:
            schema.query_type = self._healed(schema.query_type)
        if schema.mutation_type:
            schema.mutation_type = self._healed(schema.mutation_type)
        if schema.subscription_type:
            schema.subscription_type = self._healed(schema.subscription_type)

    def visit_object(self, object_definition):
        object_definition.interfaces = [
            self._healed(iface) for iface in object_definition.interfaces
        ]
        return object_definition

    def visit_field(self, field_definition):
        field_definition.type = self._healed(field_definition.type)
        return field_definition

    def visit_argument(self, argument_definition):
        argument_definition.type = self._healed(argument_definition.type)
        return argument_definition

    def visit_union(self, union):
        union.types = [self._healed(t) for t in union.types]
        return union

    def visit_input_field(self, input_field):
        input_field.type = self._healed(input_field.type)
        return input_field


class _SchemaDirectivesApplicator(SchemaVisitor):
    def __init__(self, schema, schema_directives, strict=False):
        self._schema = schema
        self._schema_directives = schema_directives
        self._strict = strict

        assert isinstance(schema, Schema)

        directive_definitions = dict(schema.directives)
        for directive_name, schema_directive in schema_directives.items():
            assert issubclass(schema_directive, SchemaDirective)
            if schema_directive.definition:
                directive_definitions[
                    directive_name
                ] = schema_directive.definition

        for directive_name in directive_definitions:
            visitor_cls = schema_directives.get(directive_name)
            if visitor_cls is None:
                continue

        self._directive_definitions = directive_definitions

    def default(self, definition):
        applied = set()
        for directive_node in _find_directives(definition):
            name = directive_node.name.value
            directive_def = self._directive_definitions.get(name)
            if directive_def is None:
                raise SDLError(
                    'Unknown directive "@%s' % name, [directive_node]
                )

            schema_directive = self._schema_directives.get(name)
            if schema_directive is None:
                if name not in _SPECIFIED_DIRECTIVE_NAMES and self._strict:
                    raise SDLError(
                        'Missing directive implementation for "@%s"' % name
                    )
                continue

            loc = _CLS_TO_LOC.get(type(definition))
            if loc not in directive_def.locations:
                raise SDLError(
                    'Directive "@%s" not applicable to "%s"' % (name, loc),
                    [directive_node],
                )

            if name in applied:
                raise SDLError(
                    'Directive "@%s" already applied' % name, [directive_node]
                )

            args = coerce_argument_values(directive_def, directive_node)
            definition = _visit(schema_directive(args), definition)
            applied.add(directive_def.name)

        return definition


# REVIEW: With the definition and the usage as a keyed map we end up repeating
# the name of the directive.
class SchemaDirective(SchemaVisitor):
    """ @directive implementation for use alongside
    :func:`py_gql.schema.build.make_executable_schema`.

    You need to subclass this in order to define your own custom directives.
    For example a directive that modifies the field resolver to always
    uppercase the result would look like this:

    .. code-block:: python

        class UppercaseDirective(SchemaDirective):

        def visit_field(self, field_definition):
            assert field_definition.type is String
            return Field(
                field_definition.name,
                field_definition.type,
                args=field_definition.args,
                description=field_definition.description,
                deprecation_reason=field_definition.deprecation_reason,
                resolve=lambda *a, **kw: field_definition.resolve(*a, **kw).upper(),
                node=field_definition.node,
            )

        # Use it as follows
        schema = make_executable_schema(
            '''
            directive @upper on FIELD_DEFINITION

            type Query {
                foo: String @upper
            }
            ''',
            schema_directives={
                'upper': UppercaseDirective
            }
        )

    Warning:
        While this is aimed to be used after
        :func:`~py_gql.schema.build.build_schema_from_ast` and its derivatives,
        schema directives can modify types inline. In case you use globally
        defined type definitions this can have some nasty side effects and so
        it is encouraged to return new type definitions instead.

    Warning:
        Specified types (scalars, introspection) cannot be modified through
        schema directives.
    """

    definition = None

    def __init__(self, args=None):
        self.args = args



# -*- coding: utf-8 -*-

from collections import namedtuple

from ..._utils import OrderedDict, deduplicate, flatten as _flatten
from ...exc import UnknownType
from ...lang import ast as _ast
from ...schema import InterfaceType, ObjectType, is_leaf_type, unwrap_type
from ...schema.types import WrappingType
from ..visitors import ValidationVisitor


def _permutations(lst):
    """ Symmetric permutations of a list

    >>> list(_permutations([1, 2, 3]))
    [(1, 2), (1, 3), (2, 3)]
    """
    for i, item_1 in enumerate(lst):
        for item_2 in lst[i + 1 :]:
            yield item_1, item_2


def _cross(iter_1, iter_2):
    """ Cross product of 2 iterables

    >>> list(_cross([1, 2, 3], [4, 5, 6]))
    [(1, 4), (1, 5), (1, 6), (2, 4), (2, 5), (2, 6), (3, 4), (3, 5), (3, 6)]
    """
    for a in iter_1:
        for b in iter_2:
            yield a, b


def _collect(iterable, target=None):
    if target is None:
        target = []

    for entry in iterable:
        target.append(entry)

    return target


def _at(lst, index, default=None):
    """
    >>> _at([], 0) is None
    True
    >>> _at([1], 2) is None
    True
    """
    try:
        return lst[index]
    except IndexError:
        return default


def _type_from_ast(schema, node):
    try:
        return schema.get_type_from_literal(node)
    except UnknownType:
        return None


Context = namedtuple(
    "Context",
    ["schema", "fields_and_fragments", "compared_fragment_pairs", "fragments"],
)


class OverlappingFieldsCanBeMergedChecker(ValidationVisitor):
    """ A selection set is only valid if all fields (including spreading any
    fragments) either correspond to distinct response names or can be merged
    without ambiguity. """

    def __init__(self, schema, type_info):
        super(OverlappingFieldsCanBeMergedChecker, self).__init__(
            schema, type_info
        )
        self.ctx = Context(self.schema, {}, set(), {})

    def enter_document(self, node):
        # Should happen before visitng any selection set
        self.ctx.fragments.update(
            {
                d.name.value: d
                for d in node.definitions
                if type(d) is _ast.FragmentDefinition
            }
        )

    def enter_selection_set(self, node):
        conflicts = find_conflicts_within_selection_set(
            self.ctx, node, self.type_info.parent_type
        )

        for response_name, reason, locs in conflicts:
            self.add_error(
                'Field(s) "%s" conflict because %s. Use different aliases on '
                "the fields to fetch both if this was intentional."
                % (response_name, reason),
                list(_flatten(locs)),
            )


def find_conflicts_within_selection_set(ctx, selection_set, parent_type):
    """ Find all conflicts found "within" a selection set, including those found
    via spreading in fragments. Called when visiting each SelectionSet in the
    GraphQL Document.


    Algorithm:
    ==========

    Conflicts occur when two fields exist in a query which will produce the same
    response name, but represent differing values, thus creating a conflict.
    The algorithm below finds all conflicts via making a series of comparisons
    between fields. In order to compare as few fields as possible, this makes a
    series of comparisons "within" sets of fields and "between" sets of fields.

    Given any selection set, a collection produces both a set of fields by
    also including all inline fragments, as well as a list of fragments
    referenced by fragment spreads.

    A) Each selection set represented in the document first compares "within"
       its collected set of fields, finding any conflicts between every pair of
       overlapping fields.
       Note: This is the *only time* that a the fields "within" a set are
       compared to each other. After this only fields "between" sets are
       compared.

    B) Also, if any fragment is referenced in a selection set, then a
       comparison is made "between" the original set of fields and the
       referenced fragment.

    C) Also, if multiple fragments are referenced, then comparisons
       are made "between" each referenced fragment.

    D) When comparing "between" a set of fields and a referenced fragment, first
       a comparison is made between each field in the original set of fields and
       each field in the the referenced set of fields.

    E) Also, if any fragment is referenced in the referenced selection set,
       then a comparison is made "between" the original set of fields and the
       referenced fragment (recursively referring to step D).

    F) When comparing "between" two fragments, first a comparison is made
       between each field in the first referenced set of fields and each field
       in the the second referenced set of fields.

    G) Also, any fragments referenced by the first must be compared to the
       second, and any fragments referenced by the second must be compared to
       the first (recursively referring to step F).

    H) When comparing two fields, if both have selection sets, then a comparison
       is made "between" both selection sets, first comparing the set of fields
       in the first selection set with the set of fields in the second.

    I) Also, if any fragment is referenced in either selection set, then a
       comparison is made "between" the other set of fields and the
       referenced fragment.

    J) Also, if two fragments are referenced in both selection sets, then a
       comparison is made "between" the two fragments.
    """
    # Implementation note: all functions detecting conflicts are generators
    # to keep the code simple and only this one collects the generators into a
    # list.
    conflicts = []

    field_map, fragment_names = _fields_and_fragments(
        ctx, parent_type, selection_set
    )

    # (A) Find find all conflicts "within" the fields of this selection set.
    _collect(_conflicts_within(ctx, field_map), conflicts)

    compared_fragments = set()
    for fragment_name in fragment_names:
        # (B) Then collect conflicts between these fields and those represented
        # by each spread fragment name found.
        _collect(
            _conflicts_between_fields_and_fragment(
                ctx, False, field_map, fragment_name, compared_fragments
            ),
            conflicts,
        )

    # (C) Then compare this fragment with all other fragments found in this
    # selection set to collect conflicts between fragments spread together.
    # This compares each item in the list of fragment names to every other
    # item in that same list (except for itself).
    for frag_1, frag_2 in _permutations(fragment_names):
        _collect(
            _conflicts_between_fragments(ctx, False, frag_1, frag_2), conflicts
        )

    return conflicts


def _fields_and_fragments(ctx, parent_type, selection_set):
    """ Given a selection set, return the collection of fields (a mapping
    of response name to field nodes and definitions) as well as a list of
    fragment names referenced via fragment spreads. """
    cached = ctx.fields_and_fragments.get(selection_set)
    if cached is not None:
        return cached

    # A field map is an ordered keyed collection, where each key represents a
    # response name and the value at that key is a list of all fields which
    # provide that response name. For every response name, if there are
    # multiple fields, they must be compared to find a potential conflict.
    field_map, fragment_names = _collect_fields_and_fragments(
        ctx, parent_type, selection_set
    )

    ctx.fields_and_fragments[selection_set] = (field_map, fragment_names)
    return field_map, list(deduplicate(fragment_names))


def _referenced_fields_and_fragments(ctx, fragment):
    """ Given a fragment definition, return the represented collection of
    fields as well as a list of nested fragment names referenced via
    fragment spreads. """
    cached = ctx.fields_and_fragments.get(fragment.selection_set)
    if cached is not None:
        return cached

    fragment_type = _type_from_ast(ctx.schema, fragment.type_condition)
    return _fields_and_fragments(ctx, fragment_type, fragment.selection_set)


def _collect_fields_and_fragments(
    ctx, parent_type, selection_set, node_and_defs=None, fragment_names=None
):
    if node_and_defs is None:
        node_and_defs = OrderedDict()
    if fragment_names is None:
        fragment_names = []

    for selection in selection_set.selections:
        kind = type(selection)
        if kind is _ast.Field:
            fieldname = selection.name.value

            fielddef = (
                parent_type.field_map.get(fieldname, None)
                if isinstance(parent_type, (ObjectType, InterfaceType))
                else None
            )

            response_name = (
                selection.alias.value
                if selection.alias and selection.alias.value
                else fieldname
            )

            if response_name not in node_and_defs:
                node_and_defs[response_name] = []

            node_and_defs[response_name].append(
                (parent_type, selection, fielddef)
            )

        elif kind is _ast.FragmentSpread:
            fragment_names.append(selection.name.value)

        elif kind is _ast.InlineFragment:
            type_condition = selection.type_condition

            inline_fragment_type = (
                _type_from_ast(ctx.schema, type_condition)
                if type_condition
                else parent_type
            )

            _collect_fields_and_fragments(
                ctx,
                inline_fragment_type,
                selection.selection_set,
                node_and_defs=node_and_defs,
                fragment_names=fragment_names,
            )

    return node_and_defs, fragment_names


def _conflicts_within(ctx, field_map):
    """ Collect all Conflicts "within" one collection of fields. """
    for response_name, fields in field_map.items():
        # This compares every field in the list to every other field in this
        # list (except to itself). If the list only has one item, nothing
        # needs to be compared.
        for field_1, field_2 in _permutations(fields):
            conflict = _find_conflict(
                ctx,
                False,  # within one collection is never mutually exclusive
                response_name,
                field_1,
                field_2,
            )
            if conflict:
                yield conflict


def _conflicts_between(
    ctx, parents_mutually_exclusive, field_map_1, field_map_2
):
    """ Collect all Conflicts between two collections of fields. This is
    similar to, but different from the `collectConflictsWithin` function above.
    This check assumes that `collectConflictsWithin` has already been called
    on each provided collection of fields. This is true because this
    validator traverses each individual selection set. """
    for response_name, fields_1 in field_map_1.items():
        fields_2 = field_map_2.get(response_name)
        if fields_2 is None:
            continue

        for field_1 in fields_1:
            for field_2 in fields_2:
                conflict = _find_conflict(
                    ctx,
                    parents_mutually_exclusive,
                    response_name,
                    field_1,
                    field_2,
                )
                if conflict:
                    yield conflict


def _conflicts_between_fields_and_fragment(
    ctx, mutually_exclusive, field_map, fragment_name, compared_fragments
):
    """ Collect all conflicts found between a set of fields and a fragment
    reference including via spreading in any nested fragments. """
    # Memoize so a fragment is not compared for conflicts more than once.
    if fragment_name in compared_fragments:
        return

    compared_fragments.add(fragment_name)
    fragment_def = ctx.fragments.get(fragment_name)
    if not fragment_def:
        return

    ff = _referenced_fields_and_fragments(ctx, fragment_def)
    fragment_field_map, fragment_fragment_names = ff

    # ??
    # Do not compare a fragment's fieldMap to itself.
    if field_map is fragment_field_map:
        return

    # (D) First collect any conflicts between the provided collection of fields
    # and the collection of fields represented by the given fragment.
    for c in _conflicts_between(
        ctx, mutually_exclusive, field_map, fragment_field_map
    ):
        yield c

    # (E) Then collect any conflicts between the provided collection of fields
    # and any fragment names found in the given fragment.
    for fragment in fragment_fragment_names:
        for c in _conflicts_between_fields_and_fragment(
            ctx, mutually_exclusive, field_map, fragment, compared_fragments
        ):
            yield c


def _conflicts_between_fragments(
    ctx, mutually_exclusive, fragment_1, fragment_2
):
    """ Collect all conflicts found between two fragments, including via
    spreading in any nested fragments. """
    if (not fragment_1) or (not fragment_2) or fragment_1 == fragment_2:
        return

    # Avoid comparing fragments twice and double reporting
    # The comparison order doesn't matter, hence the weird cache key
    cache_key = (tuple(sorted([fragment_1, fragment_2])), mutually_exclusive)
    if cache_key in ctx.compared_fragment_pairs:
        return

    ctx.compared_fragment_pairs.add(cache_key)

    def_1 = ctx.fragments.get(fragment_1)
    def_2 = ctx.fragments.get(fragment_2)

    if def_1 is None or def_2 is None:
        return

    fields_1, fragments_1 = _referenced_fields_and_fragments(ctx, def_1)
    fields_2, fragments_2 = _referenced_fields_and_fragments(ctx, def_2)

    # (F) First, collect all conflicts between these two collections of fields
    # (not including any nested fragments).
    for c in _conflicts_between(ctx, mutually_exclusive, fields_1, fields_2):
        yield c

    # (G) Then collect conflicts between the first fragment and any nested
    # fragments spread in the second fragment.
    for i, fragment in enumerate(fragments_1):
        for c in _conflicts_between_fragments(
            ctx, mutually_exclusive, fragment, _at(fragment_2, i)
        ):
            yield c

    # (G) Then collect conflicts between the second fragment and any nested
    # fragments spread in the first fragment.
    for i, fragment in enumerate(fragments_2):
        for c in _conflicts_between_fragments(
            ctx, mutually_exclusive, _at(fragment_1, i), fragment
        ):
            yield c


def _find_conflict(
    ctx, parents_mutually_exclusive, response_name, field_1, field_2
):
    """ Determines if there is a conflict between two particular fields,
    including comparing their sub-fields.
    """
    parent_1, node_1, def_1 = field_1
    parent_2, node_2, def_2 = field_2

    # If it is known that two fields could not possibly apply at the same
    # time, due to the parent types, then it is safe to permit them to diverge
    # in aliased field or arguments used as they will not present any ambiguity
    # by differing.
    # It is known that two parent types could never overlap if they are
    # different Object types. Interface or Union types might overlap - if not
    # in the current state of the schema, then perhaps in some future version,
    # thus may not safely diverge.
    mutually_exclusive = parents_mutually_exclusive or (
        parent_1 != parent_2
        and isinstance(parent_1, ObjectType)
        and isinstance(parent_2, ObjectType)
    )

    # The return type for each field.
    type_1 = def_1.type if def_1 else None
    type_2 = def_2.type if def_2 else None
    name_1 = node_1.name.value
    name_2 = node_2.name.value

    if not mutually_exclusive:
        # Two aliases must refer to the same field.
        if name_1 != name_2:
            reason = '"%s" and "%s" are different fields' % (name_1, name_2)
            return response_name, reason, (node_1, node_2)

        # Two field calls must have the same arguments.
        if not _same_arguments(node_1.arguments or [], node_2.arguments or []):
            reason = '"%s" and "%s" have different arguments' % (name_1, name_2)
            return response_name, reason, (node_1, node_2)

    if type_1 and type_2 and _types_conflict(type_1, type_2):
        reason = "they return conflicting types %s and %s" % (type_1, type_2)
        return response_name, reason, (node_1, node_2)

    # Collect and compare sub-fields. Use the same "visited fragment names" list
    # for both collections so fields in a fragment reference are never
    # compared to themselves.
    if node_1.selection_set and node_2.selection_set:
        subconflicts = _collect(
            _conflicts_between_subselections(
                ctx,
                mutually_exclusive,
                unwrap_type(type_1),
                node_1.selection_set,
                unwrap_type(type_2),
                node_2.selection_set,
            )
        )
        if subconflicts:
            reason = " and ".join(
                [
                    'subfields "%s" conflict (%s)' % (name, reason_)
                    for name, reason_, _ in subconflicts
                ]
            )
            nodes_1 = [node_1] + [nodes for _, _, (nodes, _) in subconflicts]
            nodes_2 = [node_2] + [nodes for _, _, (_, nodes) in subconflicts]
            return response_name, reason, (nodes_1, nodes_2)


def _conflicts_between_subselections(
    ctx, mutually_exclusive, parent_type_1, node_1, parent_type_2, node_2
):
    """ Find all conflicts found between two selection sets, including those
    found via spreading in fragments. Called when determining if conflicts
    exist between the sub-fields of two overlapping fields. """

    field_map_1, fragments_1 = _fields_and_fragments(ctx, parent_type_1, node_1)
    field_map_2, fragments_2 = _fields_and_fragments(ctx, parent_type_2, node_2)

    # (H) First, collect all conflicts between these two collections of field.
    for c in _conflicts_between(
        ctx, mutually_exclusive, field_map_1, field_map_2
    ):
        yield c

    # (I) Then collect conflicts between the first collection of fields and
    # those referenced by each fragment name associated with the second.
    for fragment in fragments_2:
        for c in _conflicts_between_fields_and_fragment(
            ctx, mutually_exclusive, field_map_1, fragment, set()
        ):
            yield c

    # (I) Then collect conflicts between the second collection of fields and
    # those referenced by each fragment name associated with the first.
    for fragment in fragments_1:
        for c in _conflicts_between_fields_and_fragment(
            ctx, mutually_exclusive, field_map_2, fragment, set()
        ):
            yield c

    # (J) Also collect conflicts between any fragment names by the first and
    # fragment names by the second. This compares each item in the first set of
    # names to each item in the second set of names.
    for fragment_1, fragment_2 in _cross(fragments_1, fragments_2):
        for c in _conflicts_between_fragments(
            ctx, mutually_exclusive, fragment_1, fragment_2
        ):
            yield c


def _same_arguments(args_1, args_2):
    """ Given two lists of arguments, check all arguments have the same name
    and values with no extra / missing argument.

    Args:
        args_1 (List[py_gql.lang.ast.Argument]):
        args_2 (List[py_gql.lang.ast.Argument]):

    Returns:
        bool: Whether ``args_1`` and ``args_2`` are equivalent.
    """
    if len(args_1) != len(args_2):
        return False

    s1 = sorted(args_1, key=lambda a: a.name.value)
    s2 = sorted(args_2, key=lambda a: a.name.value)

    return all(
        (
            (
                a1.name.value == a2.name.value
                and
                # noqa: E721
                type(a1.value) == type(a2.value)
                and a1.value.value == a2.value.value
            )
            for a1, a2 in zip(s1, s2)
        )
    )


def _types_conflict(type_1, type_2):
    """ Two types conflict if both types could not apply to a value
    simultaneously. Composite types are ignored as their individual field
    types will be compared later recursively. However List and Non-Null types
    must match.

     Args:
        type_1 (py_gql.schema.Type):
        type_2 (py_gql.schema.Type):

    Returns:
        bool: Whether ``type_1`` and ``type_2`` conflict.
    """
    if isinstance(type_1, WrappingType) or isinstance(type_2, WrappingType):
        return type(type_1) != type(type_2) or _types_conflict(
            type_1.type, type_2.type
        )

    if is_leaf_type(type_1) or is_leaf_type(type_2):
        return type_1 != type_2

    return False



# -*- coding: utf-8 -*-
"""
Since the June 2018 version of the specification, SDL documents are officially
supported and provide with a language agnostic way of defining schemas. The
:mod:`py_gql.schema.build` module exposes the related utilities.
"""

import collections

import six

from .. import types as _types
from ..._utils import OrderedDict, nested_key
from ...exc import ExtensionError, SDLError
from ...lang import ast as _ast, parse
from ..schema import Schema
from .type_builder import TypesBuilder
from .visitors import (
    HealSchemaVisitor,
    SchemaDirective,
    _SchemaDirectivesApplicator,
    visit_schema,
)

__all__ = (
    "build_schema_from_ast",
    "extend_schema",
    "make_executable_schema",
    "SchemaDirective",
)


def make_executable_schema(
    document,
    resolvers=None,
    additional_types=None,
    schema_directives=None,
    assume_valid=False,
):
    """ Build an executable schema from a GraphQL document.

    This includes:

    - Generating types from their definitions
    - Applying schema and type extensions
    - Applying schema directives

    Args:
        document (Union[str, py_gql.lang.ast.Document]): SDL document

        resolvers (Union[dict, callable]): Field resolvers
            If a `dict` is provided, this looks for the resolver at key
            `{type_name}.{field_name}`. If a callable is provided, this calls
            it with the `{type_name}.{field_name}` argument and use the return
            value if it is itself a callable.

        additional_types (List[py_gql.schema.Type]): User supplied list of types
            Use this to specify some custom implementation for scalar, enums,
            etc.
            - In case of object types, interfaces, etc. the supplied type will
            override the extracted type without checking for compatibility.
            - Extension will be applied to these types. As a result, the
            resulting types may not be the same objects that were provided,
            so users should not rely on type identity.

        schema_directives (dict): Schema directive classes
            - Members must be subclasses of :class:`SchemaDirective`
            - Members must define a non-null ``definition`` attribute or the
            corresponding definition must be present in the document

        assume_valid (bool): Do not validate intermediate schemas

    Returns:
        py_gql.schema.Schema: Executable schema

    Raises:
        py_gql.exc.SDLError:
    """
    ast = _document_ast(document)

    schema = build_schema_from_ast(
        ast, resolvers=resolvers, additional_types=additional_types
    )

    if not assume_valid:
        schema.validate()

    schema = extend_schema(
        schema,
        ast,
        resolvers=resolvers,
        additional_types=additional_types,
        strict=False,
    )

    if not assume_valid:
        schema.validate()

    if schema_directives:
        schema = heal_schema(
            apply_schema_directives(schema, schema_directives, strict=True)
        )

        if not assume_valid:
            schema.validate()

    return schema


def build_schema_from_ast(document, resolvers=None, additional_types=None):
    """ Build an executable schema from an SDL-based schema definition ignoring
    extensions.

    Args:
        document (Union[str, _ast.Document]): SDL document

        resolvers (Union[dict, callable]): Field resolvers
            If a `dict` is provided, this looks for the resolver at key
            `{type_name}.{field_name}`. If a callable is provided, this calls
            it with the `{type_name}.{field_name}` argument and use the return
            value if it is itself a callable.

        additional_types (List[py_gql.schema.Type]): User supplied list of types
            Use this to specify some custom implementation for scalar, enums,
            etc.
            - In case of object types, interfaces, etc. the supplied type will
            override the extracted type without checking for compatibility.
            - Extension will be applied to these types. As a result, the
            resulting types may not be the same objects that were provided,
            so users should not rely on type identity.

    Returns:
        py_gql.schema.Schema: Executable schema

    Raises:
        py_gql.exc.SDLError:
    """
    ast = _document_ast(document)
    schema_def, type_defs, directive_defs = _collect_definitions(ast)

    builder = TypesBuilder(
        type_defs,
        directive_defs,
        additional_types=_merge_type_maps(additional_types),
    )

    directives = [
        builder.build_directive(directive_def)
        for directive_def in directive_defs.values()
    ]

    types = [builder.build_type(type_def) for type_def in type_defs.values()]

    _assign_resolvers(types, resolvers)

    if schema_def is None:
        operations = {
            t.name.lower(): t
            for t in types
            if t.name in ("Query", "Mutation", "Subscription")
        }
    else:
        operations = {}
        for op_def in schema_def.operation_types:
            op = op_def.operation
            if op in operations:
                raise SDLError(
                    "Schema must only define a single %s operation" % op,
                    [schema_def, op_def],
                )
            operations[op] = builder.build_type(op_def.type)

    return Schema(
        query_type=operations.get("query"),
        mutation_type=operations.get("mutation"),
        subscription_type=operations.get("subscription"),
        types=types,
        directives=directives,
        nodes=[schema_def] if schema_def else None,
    )


def extend_schema(
    schema, document, resolvers=None, additional_types=None, strict=True
):
    """ Extend an existing Schema according to a GraphQL document (adding new
    types and directives + extending known types).

    Warning:

        Specified types cannot be replace or extended.

    Args:
        schema (py_gql.schema.Schema): Executable schema

        document (Union[str, _ast.Document]): SDL document

        resolvers (Union[dict, callable]): Field resolvers
            If a `dict` is provided, this looks for the resolver at key
            `{type_name}.{field_name}`. If a callable is provided, this calls
            it with the `{type_name}.{field_name}` argument and use the return
            value if it is itself a callable.

        additional_types (List[py_gql.schema.Type]): User supplied list of types
            Use this to specify some custom implementation for scalar, enums,
            etc.
            - In case of object types, interfaces, etc. the supplied type will
            override the extracted type without checking for compatibility.
            - Extension will be applied to these types. As a result, the
            resulting types may not be the same objects that were provided,
            so users should not rely on type identity.

        strict (bool): Enable strict mode.
            In strict mode, unknown extension targets, overriding type and
            overriding the schema definition will raise an
            :class:`ExtensionError`. Disable strict mode will silently ignore
            such errors.

    Returns:
        py_gql.schema.Schema: Executable schema

    Raises:
        py_gql.exc.SDLError:

    """
    ast = _document_ast(document)
    assert isinstance(schema, Schema)

    schema_exts, type_defs, directive_defs, type_exts = _collect_extensions(
        schema, ast, strict=strict
    )

    if not (schema_exts or type_defs or directive_defs or type_exts):
        return schema

    if additional_types is None:
        additional_types = {}

    builder = TypesBuilder(
        type_defs,
        directive_defs,
        type_exts,
        additional_types=_merge_type_maps(schema.types, additional_types),
    )

    directives = [
        builder.extend_directive(d) for d in schema.directives.values()
    ]
    directives += [
        builder.extend_directive(builder.build_directive(d))
        for d in directive_defs.values()
    ]

    types = [builder.extend_type(t) for t in schema.types.values()]
    types += [
        builder.extend_type(builder.build_type(t)) for t in type_defs.values()
    ]

    _assign_resolvers(types, resolvers)

    def _extend_or(maybe_type):
        return builder.extend_type(maybe_type) if maybe_type else None

    operation_types = dict(
        query=_extend_or(schema.query_type),
        mutation=_extend_or(schema.mutation_type),
        subscription=_extend_or(schema.subscription_type),
    )

    for ext in schema_exts:
        for op_def in ext.operation_types:
            op = op_def.operation
            if operation_types.get(op) is not None:
                raise ExtensionError(
                    "Schema must only define a single %s operation" % op,
                    [ext, op_def],
                )
            operation_types[op] = builder.extend_type(
                builder.build_type(op_def.type)
            )

    return Schema(
        query_type=operation_types.get("query", None),
        mutation_type=operation_types.get("mutation", None),
        subscription_type=operation_types.get("subscription", None),
        types=types,
        directives=directives,
        nodes=(schema.nodes or []) + (schema_exts or []),
    )


def _collect_definitions(document):
    schema_definition = None
    types = OrderedDict()
    directives = OrderedDict()

    for definition in document.definitions:
        if isinstance(definition, _ast.SchemaDefinition):
            if schema_definition is not None:
                raise SDLError(
                    "Must provide only one schema definition", [definition]
                )
            schema_definition = definition

        elif isinstance(definition, _ast.TypeDefinition):
            if definition.name.value in types:
                raise SDLError(
                    "Duplicate type %s" % definition.name.value, [definition]
                )
            types[definition.name.value] = definition

        elif isinstance(definition, _ast.DirectiveDefinition):
            if definition.name.value in directives:
                raise SDLError(
                    "Duplicate directive @%s" % definition.name.value,
                    [definition],
                )
            directives[definition.name.value] = definition

    return schema_definition, types, directives


def _assign_resolvers(types, resolvers):

    if callable(resolvers):
        infer = lambda parent, field: resolvers(parent.name, field.name)
    elif isinstance(resolvers, dict):
        infer = lambda parent, field: (
            resolvers.get("%s.%s" % (parent.name, field.name), None)
            or nested_key(resolvers, parent.name, field.name, default=None)
        )
    elif not resolvers:
        return

    for type_ in types:
        if isinstance(type_, _types.ObjectType):
            for field in type_.fields:
                field.resolve = infer(type_, field) or field.resolve


def _document_ast(document):
    # type: (Union[_ast.Document, str]) -> _ast.Document
    if isinstance(document, six.string_types):
        return parse(document, allow_type_system=True)
    elif isinstance(document, _ast.Document):
        return document
    else:
        TypeError("Expected Document but got %s" % type(document))


def _collect_extensions(schema, document, strict=True):
    schema_exts = []
    type_defs = OrderedDict()
    _type_exts = []
    type_exts = collections.defaultdict(list)
    directive_defs = OrderedDict()

    for definition in document.definitions:
        if strict and isinstance(definition, _ast.SchemaDefinition):
            raise ExtensionError(
                "Cannot redefine schema in strict schema extension",
                [definition],
            )

        elif isinstance(definition, _ast.SchemaExtension):
            schema_exts.append(definition)

        elif isinstance(definition, _ast.TypeDefinition):
            name = definition.name.value
            existing = schema.get_type(name, None)
            if existing is not None:
                if strict:
                    raise ExtensionError(
                        'Type "%s" is already defined in the schema.' % name,
                        [definition],
                    )
                else:
                    continue
            type_defs[name] = definition

        elif isinstance(definition, _ast.DirectiveDefinition):
            name = definition.name.value
            existing = schema.directives.get(name, None)
            if existing is not None:
                if strict:
                    raise ExtensionError(
                        'Directive "@%s" is already defined in the schema.'
                        % name,
                        [definition],
                    )
                else:
                    continue
            directive_defs[name] = definition

        elif isinstance(definition, _ast.TypeExtension):
            _type_exts.append(definition)

    for ext in _type_exts:
        target = ext.name.value
        target_exists = schema.get_type(target, None) or (target in type_defs)
        if not target_exists:
            if strict:
                raise ExtensionError(
                    'Cannot extend undefined type "%s".' % target, [ext]
                )
            else:
                continue

        type_exts[target].append(ext)

    return schema_exts, type_defs, directive_defs, type_exts


def _merge_type_maps(*type_maps):
    type_map = {}
    for tm in type_maps:
        if isinstance(tm, list):
            type_map.update({type_.name: type_ for type_ in tm})
        elif isinstance(tm, dict):
            type_map.update(tm)
    return type_map


def heal_schema(schema):
    """ Fix type reference in a schema after modifying them inline.

    This ensures that all nested references to a certain type match the top
    level reference the schema has.

    Args:
        schema (py_gql.schema.Schema): Schema to fix

    Returns:
        py_gql.schema.Schema: Fixed schema

    Warning:
        This can modify the types inline and is expected to be used after
        buiding a schema, do not use this if your types are globally defined.
    """
    return visit_schema(HealSchemaVisitor(schema), schema)


def apply_schema_directives(schema, schema_directives, strict=False):
    """ Apply schema directives to a schema.

    Args:
        schema (py_gql.schema.Schema): Schema to modify
        schema_directives (Mapping[str, type]): ``{ name -> SchemaDirective subclass }``
            Schema directives are instantiated and provided the arguments
            for each occurence so user need to provide classes.
        strict (bool):
            If ``True`` will raise on missing implementation, otherwise silently
            ignores such directivess

    Returns:
        py_gql.schema.Schema: Updated schema
    """
    return visit_schema(
        _SchemaDirectivesApplicator(schema, schema_directives, strict=strict),
        schema,
    )



# -*- coding: utf-8 -*-


def default_resolver(parent_value, args, context, info):
    """ Default resolver used during query execution.

    This resolver looks up the value from the ``parent_value`` in the
    following lookup order:

    1. If ``parent_value`` is a dict subcass:

        1. If the field name is present and non callable, return it
        2. If the field name is present and callable, return the result of
           calling it like a normal resolver (i.e. with the same arguments
           that were passed to this function)

    2. If ``parent_value`` has an attribute corresponding to the field name:

        1.  If the attribute is non callable, return it
        2. If the attribute is callable, treat it like a method of
           ``parent_value`` and call it passing in ``(args, context, info)``

    3. Return ``None``

    Args:
        parent_value (any): Value of the resolved parent node
        args (dict): Coerced field arguments
        context (any): User provided context value
        info (class:`py_gql.execution.ResolveInfo`): Resolution context

    Returns:
        any: Resolved value
    """
    field_name = info.field_def.name
    try:
        field_value = parent_value.get(field_name, None)
    except AttributeError:
        attr_value = getattr(parent_value, field_name, None)
        if callable(attr_value):
            return attr_value(args, context, info)
        return attr_value
    else:
        if callable(field_value):
            return field_value(parent_value, args, context, info)
        return field_value



# -*- coding: utf-8 -*-
""" Validation of GraphQL (query) documents.

Note:
    This module is only concerned with validating query documents, not SDL
    documents.

.. autoattribute:: py_gql.validation.SPECIFIED_RULES
    :annotation:

    This is the list of :class:`~py_gql.validation.ValidationVisitor`
    from :mod:`py_gql.validation.rules` encoding all the validation rules
    defined in `this section
    <http://facebook.github.io/graphql/June2018/#sec-Validation>`_ of the Spec.
"""

# flake8: noqa

from .validate import SPECIFIED_RULES, ValidationResult, validate_ast
from .visitors import ValidationVisitor, VariablesCollector

__all__ = (
    "ValidationResult",
    "validate_ast",
    "SPECIFIED_RULES",
    "ValidationVisitor",
    "VariablesCollector",
)



# -*- coding: utf-8 -*-

import logging
import time

from py_gql import graphql
from py_gql.schema import Field, ObjectType, Schema, String
from py_gql.utilities.tracers import SlowQueryLog


def slow_resolver(*args, **kwargs):
    time.sleep(0.2)
    return "Foo"


def test_slow_query_log(caplog):
    schema = Schema(
        ObjectType("Query", [Field("foo", String, resolve=slow_resolver)])
    )

    graphql(schema, "{ foo }", tracer=SlowQueryLog(100))

    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelno == logging.WARNING
    assert record.name == "py_gql.utilities.tracers.SlowQueryLog"
    _, _, op, doc, variables = record.args
    assert op is None
    assert doc == "{ foo }"
    assert variables == "null"


def test_slow_query_log_not_triggered(caplog):
    schema = Schema(
        ObjectType("Query", [Field("foo", String, resolve=slow_resolver)])
    )
    graphql(schema, "{ foo }", tracer=SlowQueryLog(3000))
    assert len(caplog.records) == 0


def test_slow_query_log_custom_logger(caplog):
    schema = Schema(
        ObjectType("Query", [Field("foo", String, resolve=slow_resolver)])
    )
    logger = logging.getLogger("my_custom_logger")
    logger.setLevel(logging.DEBUG)

    graphql(
        schema,
        "{ foo }",
        tracer=SlowQueryLog(100, logger=logger, level=logging.INFO),
    )

    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelno == logging.INFO
    assert record.name == "my_custom_logger"


def test_slow_query_log_custom_class(caplog):
    class RedactedLogger(SlowQueryLog):
        def format_document(self, document):
            return "<REDACTED>"

        def format_variables(self, document):
            return "<REDACTED>"

    schema = Schema(
        ObjectType("Query", [Field("foo", String, resolve=slow_resolver)])
    )
    logger = logging.getLogger("my_custom_logger")
    logger.setLevel(logging.DEBUG)

    graphql(
        schema,
        "{ foo }",
        tracer=RedactedLogger(100, logger=logger, level=logging.INFO),
    )

    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelno == logging.INFO
    assert record.name == "my_custom_logger"

    _, _, _, doc, variables = record.args
    assert doc == "<REDACTED>"
    assert variables == "<REDACTED>"
